/*
 * Copyright (C) 2012 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Pawel Stolowski <pawel.stolowski@canonical.com>
 *
 */

namespace Unity.Protocol
{

  private static const string SCOPES_DIR = "unity/scopes";

  errordomain ParseError
  {
    INVALID_METADATA,
    FILE_NOT_FOUND,
    INVALID_PATH,
    UNKNOWN_FILE
  }

  public class MetaDataColumnInfo
  {
    public string name { get; internal set; }
    public string type_id { get; internal set; }

    /**
     * Creates MetaDataColumnInfo from a string in format "name[t]", where 't' is GVariant type.
     */
    public MetaDataColumnInfo (string v) throws Error
    {
      int idx = v.index_of ("[");
      if (idx > 0 && v.index_of ("]") == v.length - 1) //there must be at least 1 character preceding [, thus > 0
      {
        name = v.substring (0, idx);
        type_id = v.slice (idx + 1, v.length - 1);
      }
      else
      {
        throw new ParseError.INVALID_METADATA ("Invalid format of meta data string");
      }
    }

    public virtual bool equals (MetaDataColumnInfo m)
    {
      return name == m.name && type_id == m.type_id;
    }
  }

  public class MetaDataSchemaInfo
  {
    public GLib.GenericArray<MetaDataColumnInfo> columns;

    /**
     * Creates a list of MetaDataColumnInfo objects from a semicolon-separated strings, e.g. "name1[type];name2[type2]...".
     *
     * @param mdinfo_str semicolon-separated name[type] entries
     * @return list of MetaDataColumnInfo objects
     */
    public static MetaDataSchemaInfo from_string (string mdinfo_str) throws Error
    {
      var schema_info = new MetaDataSchemaInfo ();
      schema_info.columns = new GLib.GenericArray<MetaDataColumnInfo> ();

      foreach (var m in mdinfo_str.split (";"))
      {
        if (m.length > 0)
        {
          var info = new MetaDataColumnInfo (m);
          schema_info.columns.add (info);
        }
      }
      return schema_info;
    }

    public HashTable<string, string> as_hash_table ()
    {
      HashTable<string, string> ret = new HashTable<string, string> (str_hash, str_equal);

      for (int i = 0; i < columns.length; i++)
      {
        unowned MetaDataColumnInfo info = columns[i];
        ret[info.name] = info.type_id;
      }

      return ret;
    }

    public bool equals (MetaDataSchemaInfo s)
    {
      if (columns.length != s.columns.length)
        return false;
      for (int i = 0; i < columns.length; i++)
      {
        if (!columns[i].equals (s.columns[i]))
          return false;
      }
      return true;
    }
  }

  public class ScopeRegistry
  {
    private static const string SCOPE_GROUP = "Scope";

    public class ScopeMetadata
    {
      public string id;
      public string full_path;
      public string name;
      public string dbus_path;
      public string dbus_name;
      public string icon;
      public string category_icon;
      public MetaDataSchemaInfo? required_metadata;
      public MetaDataSchemaInfo? optional_metadata;
      public GLib.SList<string> keywords;
      public string type;
      public string description;
      public string search_hint;
      public bool is_master;
      public bool global_searches;
      public bool visible;
      public bool remote_content;
      public string query_binary;
      public string query_pattern;
      public string shortcut;

      public void load_from_key_file (KeyFile file) throws Error
      {
        // required fields
        this.name = file.get_string (SCOPE_GROUP, "Name");
        this.dbus_name = file.get_string (SCOPE_GROUP, "DBusName");
        this.dbus_path = file.get_string (SCOPE_GROUP, "DBusPath");
        this.icon = file.get_string (SCOPE_GROUP, "Icon");
        this.type = file.get_string (SCOPE_GROUP, "Type");

        // optional fields
        if (file.has_key (SCOPE_GROUP, "IsMaster"))
          this.is_master = file.get_boolean (SCOPE_GROUP, "IsMaster");
        else
          this.is_master = false;

        if (file.has_key (SCOPE_GROUP, "Visible"))
          this.visible = file.get_boolean (SCOPE_GROUP, "Visible");
        else
          this.visible = true;

        if (file.has_key (SCOPE_GROUP, "GlobalSearches"))
          this.global_searches = file.get_boolean (SCOPE_GROUP, "GlobalSearches");
        else
          this.global_searches = false;

        if (file.has_key (SCOPE_GROUP, "RemoteContent"))
          this.remote_content = file.get_boolean (SCOPE_GROUP, "RemoteContent");
        else
          this.remote_content = false;

        if (file.has_key (SCOPE_GROUP, "QueryBinary"))
          this.query_binary = file.get_string (SCOPE_GROUP, "QueryBinary");

        if (file.has_key (SCOPE_GROUP, "CategoryIcon"))
          this.category_icon = file.get_string (SCOPE_GROUP, "CategoryIcon");

        if (file.has_key (SCOPE_GROUP, "QueryPattern"))
          this.query_pattern = file.get_string (SCOPE_GROUP, "QueryPattern");

        if (file.has_key (SCOPE_GROUP, "Description"))
          this.description = file.get_string (SCOPE_GROUP, "Description");

        if (file.has_key (SCOPE_GROUP, "SearchHint"))
          this.search_hint = file.get_string (SCOPE_GROUP, "SearchHint");

        if (file.has_key (SCOPE_GROUP, "RequiredMetadata"))
          this.required_metadata = MetaDataSchemaInfo.from_string (file.get_string (SCOPE_GROUP, "RequiredMetadata"));

        if (file.has_key (SCOPE_GROUP, "OptionalMetadata"))
          this.optional_metadata = MetaDataSchemaInfo.from_string (file.get_string (SCOPE_GROUP, "OptionalMetadata"));

        if (file.has_key (SCOPE_GROUP, "Keywords"))
        {
          // split keywords
          this.keywords = new GLib.SList<string> ();
          foreach (var k in file.get_string (SCOPE_GROUP, "Keywords").split (";"))
          {
            if (k.length > 0)
              this.keywords.append (k);
          }
        }

        if (file.has_key (SCOPE_GROUP, "Shortcut"))
        {
          this.shortcut = file.get_string (SCOPE_GROUP, "Shortcut");
        }
      }

      public static ScopeMetadata for_id (string scope_id) throws Error
      {
        debug ("for_id: %s", scope_id);

        string full_path;
        var file = new KeyFile ();
        var path = "%s/%s".printf (SCOPES_DIR, scope_id);
        bool loaded = file.load_from_data_dirs (path, out full_path,
                                                KeyFileFlags.NONE);

        if (!loaded)
            throw new ParseError.FILE_NOT_FOUND ("Scope not found: %s", scope_id);

        ScopeMetadata data = new ScopeMetadata ();
        data.id = scope_id;
        data.full_path = full_path;

        data.load_from_key_file (file);

        return data;
      }
    }

    // node representing single scope, with optional sub-scopes (if master scope)
    public class ScopeRegistryNode
    {
      public ScopeMetadata scope_info;
      public GLib.SList<ScopeMetadata?>? sub_scopes;
    }

    // list of top-level scopes
    private SList<ScopeRegistryNode> scopes_ = new SList<ScopeRegistryNode> ();
    public GLib.SList<ScopeRegistryNode>? scopes
    {
      get { return scopes_; }
    }

    private ScopeRegistry ()
    {
    }

    internal static async ScopeMetadata? parse_scope_file (string path, string? master_prefix) throws Error
    {
      debug ("parse_scope_file: %s, master_prefix = %s", path, master_prefix);

      bool loaded = true;
      ScopeMetadata data = new ScopeMetadata ();

      var file = new KeyFile ();
      if (GLib.Path.is_absolute (path))
      {
        loaded = file.load_from_file (path, KeyFileFlags.NONE);
        data.full_path = path;
      }
      else
      {
        loaded = file.load_from_data_dirs (path, out data.full_path, KeyFileFlags.NONE);
      }

      if (!loaded)
        throw new ParseError.FILE_NOT_FOUND ("File not found: %s", path);

      data.load_from_key_file (file);

      var basename = GLib.Path.get_basename (path);
      data.id = basename;
      if (master_prefix != null && master_prefix.length > 0)
        data.id = "%s-%s".printf (master_prefix, basename);

      debug ("parse_scope_file: %s finished", path);
      return data;
    }

    /**
     * Process .scope file and scopes in a subdirectory if it's master scope.
     *
     * @param filename path of a .scope file
     */
    private async void process_scope (string filename, string? master_scope_prefix = null) throws Error
    {
      debug ("process_scope: %s", filename);

      ScopeMetadata? scope_data = null;
      scope_data = yield parse_scope_file (filename, master_scope_prefix); // this may throw, in such case don't process this scope (and possibly its subscopes if it was master scope)

      if (scope_data != null)
      {
        var scope_node = new ScopeRegistryNode ()
        {
          scope_info = scope_data,
          sub_scopes = null
        };

        if (scope_data.is_master)
        {
          // check if there is a subdirectory with the name of master scope.
          // note that failure on single sub-scope shouldn't disable master & valid sub-scopes.
          var scopefile = GLib.File.new_for_path (scope_data.full_path);
          string scope_name = remove_scope_extension (scope_data.id);
          var parent = scopefile.get_parent ();
          if (parent != null)
          {
            string check_path = Path.build_filename (parent.get_path (), scope_name);

            if (FileUtils.test (check_path, FileTest.IS_DIR))
            {
              GLib.Dir? scope_subdir = null;
              try
              {
                scope_subdir = GLib.Dir.open (check_path);
              }
              catch (FileError e)
              {
                warning ("Error accessing directory %s", check_path);
              }

              if (scope_subdir != null)
              {
                // iterate over all files, find .*scope
                string slave_scope_filename;
                while ((slave_scope_filename = scope_subdir.read_name ()) != null)
                {
                  slave_scope_filename = Path.build_filename (check_path, slave_scope_filename);
                  ScopeMetadata? slave_scope_data = null;
                  try
                  {
                    slave_scope_data = yield parse_scope_file (slave_scope_filename, scope_name);
                  }
                  catch (Error e)
                  {
                    warning ("Failed to process '%s': %s", slave_scope_filename, e.message);
                  }
                  if (slave_scope_data != null)
                  {
                    if (scope_node.sub_scopes == null)
                      scope_node.sub_scopes = new GLib.SList<ScopeMetadata?> ();

                    scope_node.sub_scopes.append (slave_scope_data);
                  }
                }
              }
            }
          }
        }

        scopes_.append (scope_node);
      }
    }

    /**
     * Build registry of all scopes in start_path.
     * start_path can be a directory, a .scope file path or just scope id (.scope file name, including extenstion).
     *
     * @param start_path starting directory or specific .scope file
     * @return registry of all scopes (if start_path is a dir) or just one scope and its subscopes.
     */
    public static async ScopeRegistry find_scopes (string start_path) throws Error
    {
      var reg = yield find_scopes_for_master (start_path, null);
      return reg;
    }

    /**
     * Build registry of all scopes in start_path.
     * start_path can be a directory, a .scope file path or just scope id (.scope file name, including extenstion).
     *
     * @param start_path starting directory or specific .scope file
     * @return registry of all scopes (if start_path is a dir) or just one scope and its subscopes.
     */
    private static async ScopeRegistry find_scopes_for_master (string start_path, string? master_prefix) throws Error
    {
      debug ("find_scopes_for_master: %s\n", start_path);

      var reg = new ScopeRegistry ();

      if (FileUtils.test (start_path, FileTest.IS_DIR))
      {
        var dir = GLib.Dir.open (start_path);
        string name;

        while ((name = dir.read_name ()) != null)
        {
          string filename = Path.build_filename (start_path, name);
          if (filename.has_suffix (".scope"))
          {
            debug ("Found scope file: %s", filename);
            try // failure of single scope shouldn't break processing of others scopes
            {
              yield reg.process_scope (filename, master_prefix);
            }
            catch (Error e)
            {
              warning ("Failed to process '%s': %s", filename, e.message);
            }
          }
        }
      }
      else
      {
        if (start_path.has_suffix (".scope"))
        {
          yield reg.process_scope (start_path); // no try-catch here, as request for specific scope file should fail on error
        }
        else
        {
          throw new ParseError.UNKNOWN_FILE ("Unknown file type");
        }
      }
      debug ("find_scopes_for_master finished");
      return reg;
    }

    internal static string remove_scope_extension (string scope_id)
    {
      if (scope_id.has_suffix (".scope"))
        return scope_id.substring (0, scope_id.last_index_of ("."));
      return scope_id;
    }

    /**
     * Find sub-scopes for given master scope id in unity/scopes subdirectory
     * of XDG_DATA_DIRS dirs or in root_path.
     * @param scope_id id of a master scope (with or without .scope suffix)
     * @param root_path base directory of scopes, defaults to XDG_DATA_DIRS paths + "/unity/scopes"
     * @return scope registry with scopes property populated with all sub-scopes of the master scope.
     */
    public static async ScopeRegistry find_scopes_for_id (string scope_id, string? root_path = null) throws Error
    {
      debug ("find_scopes_for_id: %s", scope_id);

      string[]? dirs = null;
      if (root_path == null)
      {
        dirs = GLib.Environment.get_system_data_dirs ();
      }
      else
      {
        dirs = new string [1] {root_path};
      }

      if (dirs == null || dirs.length == 0)
      {
        throw new ParseError.INVALID_PATH ("Invalid scopes path");
      }

      var sc = remove_scope_extension (scope_id);
      foreach (var path in dirs)
      {
        var check_path = root_path == null ? Path.build_path (Path.DIR_SEPARATOR_S, path, SCOPES_DIR, sc) : Path.build_path (Path.DIR_SEPARATOR_S, root_path, sc);
        if (FileUtils.test (check_path, FileTest.IS_DIR))
        {
          var reg = yield find_scopes_for_master (check_path, sc);
          return reg;
        }
      }

      debug ("No sub-scopes directory found for master scope %s", scope_id);
      return new ScopeRegistry ();
    }
  }
}  /* namespace unity */
