/*
 * Copyright (C) 2011-2012 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 *             Michal Hruby <michal.hruby@canonical.com>
 *
 */

using GLib;
using Dee;
using Unity;
using Unity.Protocol;

namespace Unity.Internal {

private class AggregatorScopeImpl : GLib.Object, ScopeService, ScopeDBusImpl,
                                    DeprecatedScopeDBusImpl, MergeStrategy
{
  private static Quark dedup_model_quark = Quark.from_string ("unity-dedup-model");

  private class ScopeTracker : Object
  {
    private GenericArray<ScopeProxy> scope_proxy_arr;
    // scope_id -> ScopeProxy
    private HashTable<string, Utils.AsyncOnce<ScopeProxy>> scope_proxies;
    // scope_uid -> ScopeProxy
    private HashTable<string, ScopeProxy> scope_proxy_uids;
    // channel_key (proxy + master_channel_id) -> scope channel_id
    private HashTable<string, Utils.AsyncOnce<string>> scope_channel_ids;
    // channel_key (proxy + master_channel_id) -> Dee.Model
    private HashTable<string, Dee.SerializableModel> scope_models;
    // master_channel_id -> ResultsSynchronizer
    private HashTable<string, ResultsSynchronizer> synchronizers;
    // scope_id -> RemoteContent
    private HashTable<string, bool> scope_has_remote;

    public ScopeTracker ()
    {
      scope_proxy_arr = new GenericArray<ScopeProxy> ();
      scope_proxies = new HashTable<string, Utils.AsyncOnce<ScopeProxy>> (str_hash, str_equal);
      scope_proxy_uids = new HashTable<string, ScopeProxy> (str_hash,str_equal);
      scope_channel_ids = new HashTable<string, Utils.AsyncOnce<string>> (str_hash, str_equal);
      scope_models = new HashTable<string, Dee.SerializableModel> (str_hash, str_equal);
      synchronizers = new HashTable<string, ResultsSynchronizer> (str_hash, str_equal);
      scope_has_remote = new HashTable<string, bool> (str_hash, str_equal);
    }

    public List<weak string> scope_ids_for_proxies ()
    {
      return scope_proxies.get_keys ();
    }

    public string? scope_id_for_proxy (ScopeProxy proxy)
    {
      foreach (unowned string scope_id in scope_proxies.get_keys ())
      {
        if (scope_proxies[scope_id].get_data () == proxy) return scope_id;
      }

      return null;
    }

    public unowned ResultsSynchronizer? get_synchronizer (string channel_id)
    {
      return synchronizers.lookup (channel_id);
    }

    private bool content_enabled (bool scope_contains_remote)
    {
      var pref_man = PreferencesManager.get_default ();
      if (scope_contains_remote && pref_man.remote_content_search == PreferencesManager.RemoteContent.NONE)
      {
        return false;
      }
      return true;
    }

    public async ScopeProxy? create_proxy (ScopeRegistry.ScopeMetadata metadata)
    {
      ScopeProxy? proxy = null;
      try
      {
        proxy = yield ScopeProxy.new_from_metadata (metadata);
        // check that the proxy props match the metadata
        check_proxy_vs_metadata (proxy, metadata);
        scope_proxy_arr.add (proxy);

        proxy.channels_invalidated.connect (channels_invalidated);
      }
      catch (Error err)
      {
        warning ("Unable to acquire proxy for scope \"%s\": %s",
                 metadata.id, err.message);
      }
      return proxy;
    }

    private void channels_invalidated (ScopeProxy proxy)
    {
      proxy_disconnected (proxy);

      // invalidate all associated containers
      string[] invalid_keys = {};
      string prefix = "%p::".printf (proxy);
      
      debug ("Invalidating channels for %s", prefix);

      foreach (unowned string channel_key in scope_channel_ids.get_keys ())
      {
        if (channel_key.has_prefix (prefix)) invalid_keys += channel_key;
      }

      foreach (unowned string channel_key in invalid_keys)
      {
        scope_channel_ids.remove (channel_key);
        scope_models.remove (channel_key);
      }
    }

    private void check_proxy_vs_metadata (
        ScopeProxy proxy,
        ScopeRegistry.ScopeMetadata metadata) throws Error
    {
      if (proxy.is_master != metadata.is_master)
        throw new ScopeError.DATA_MISMATCH ("Scope file info for '%s' doesn't match on IsMaster key".printf (metadata.id));

      if (metadata.required_metadata != null)
      {
        var dict = metadata.required_metadata.as_hash_table ();
        unowned string field_name;
        unowned string schema;
        var iter = HashTableIter<string, string> (dict);
        while (iter.next (out field_name, out schema))
        {
          if (proxy.metadata[field_name] != schema)
            throw new ScopeError.DATA_MISMATCH ("Scope file info for '%s' doesn't match on RequiredMetadata key".printf (metadata.id));
        }
      }
    }

    private string get_channel_key (string master_channel_id, ScopeProxy proxy)
    {
      return "%p::%s".printf (proxy, master_channel_id);
    }

    public void register_channel (string master_channel_id,
                                  Dee.SerializableModel model,
                                  MergeStrategy merge_strategy)
    {
      // create new synchronizer for this channel
      var synchronizer = new ResultsSynchronizer (model);
      synchronizer.merge_strategy = merge_strategy;
      synchronizers[master_channel_id] = synchronizer;
    }

    public void unregister_channel (string master_channel_id)
    {
      var synchronizer = synchronizers[master_channel_id];
      if (synchronizer != null)
      {
        // break the circular reference
        synchronizer.receiver.set_qdata<Dee.Model?> (dedup_model_quark, null);
        synchronizers.remove (master_channel_id);
      }
      // FIXME: close child scopes' channels
    }
    
    private void proxy_disconnected (ScopeProxy proxy)
    {
      foreach (var scope_id in scope_proxies.get_keys ())
      {
        var data = scope_proxies[scope_id].get_data ();
        if (data == proxy)
        {
          warning ("Lost proxy connection to %s", scope_id);

          string prefix = "%p:".printf (proxy);
          foreach (var channel_key in scope_channel_ids.get_keys ())
          {
            if (channel_key.has_prefix (prefix))
            {
              string master_id = channel_key.substring (channel_key.last_index_of ("::") + 2);
              if (synchronizers.contains (master_id))
              {
                debug ("Removing %s data from model of master channel %s", scope_id, master_id);
                synchronizers[master_id].clear_provider_model (prefix);
                synchronizers[master_id].remove_provider (prefix);  // isn't clear+remove racy here?
              }
            }
          }

          break;
        }
      }
    }
    
    private async void wait_for_seqnum (Dee.SharedModel model, uint64 seqnum)
    {
      if (model.get_seqnum () >= seqnum) return;

      ulong update_sig_id = 0;
      update_sig_id = model.end_transaction.connect ((m, begin_seqnum, end_seqnum) =>
      {
        if (end_seqnum < seqnum) return;

        /* disconnect from within the signal handler... awesome, right? */
        SignalHandler.disconnect (m, update_sig_id);
        wait_for_seqnum.callback ();
      });
      yield;
    }

    public unowned string get_child_scope_channel_id (string master_channel_id,
                                                      ScopeProxy proxy)
    {
      var channel_key = get_channel_key (master_channel_id, proxy);
      return scope_channel_ids[channel_key].get_data ();
    }

    private Utils.AsyncOnce<ScopeProxy> get_proxy_once (string scope_id)
    {
      var proxy_once = scope_proxies[scope_id];
      if (proxy_once == null)
      {
        proxy_once = new Utils.AsyncOnce<ScopeProxy> ();
        scope_proxies[scope_id] = proxy_once;
      }

      return proxy_once;
    }

    private Utils.AsyncOnce<string> get_channel_id_once (string channel_key)
    {
      var channel_id_once = scope_channel_ids[channel_key];
      if (channel_id_once == null)
      {
        channel_id_once = new Utils.AsyncOnce<string> ();
        scope_channel_ids[channel_key] = channel_id_once;
      }

      return channel_id_once;
    }

    private unowned string? get_channel_id (string master_channel_id,
                                            string scope_id,
                                            out ScopeProxy? proxy) throws Error
    {
      var proxy_once = get_proxy_once (scope_id);
      if (!proxy_once.is_initialized ())
      {
        proxy = null;
        return null;
      }

      if (proxy_once.get_data () == null)
        throw new ScopeError.REQUEST_FAILED ("Unable to create proxy");

      proxy = proxy_once.get_data ();

      var channel_key = get_channel_key (master_channel_id, proxy);
      var channel_id_once = get_channel_id_once (channel_key);
      if (!channel_id_once.is_initialized ()) return null;
      return channel_id_once.get_data ();
    }

    private async unowned string init_channel (string master_channel_id,
                                               string scope_id,
                                               ChannelType channel_type,
                                               out ScopeProxy proxy)
      throws Error
    {
      // init ScopeProxy
      var proxy_once = get_proxy_once (scope_id);
      Error? failure = null;
      bool ignore_proxy = false;
      if (!proxy_once.is_initialized ())
      {
        if (yield proxy_once.enter ())
        {
          ScopeProxy? actual_proxy = null;
          try
          {
            // don't even create the proxy if its content is disabled
            if (scope_id in scope_has_remote)
            {
              if (!content_enabled (scope_has_remote[scope_id]))
              {
                ignore_proxy = true;
                throw new ScopeError.DATA_MISMATCH ("Remote searches disabled");
              }
            }
            var metadata = ScopeRegistry.ScopeMetadata.for_id (scope_id);
            scope_has_remote[scope_id] = metadata.remote_content;
            if (!content_enabled (scope_has_remote[scope_id]))
            {
              ignore_proxy = true;
              throw new ScopeError.DATA_MISMATCH ("Remote searches disabled");
            }
            actual_proxy = yield create_proxy (metadata);
          }
          catch (Error e)
          {
            failure = e;
          }
          finally
          {
            proxy_once.leave (actual_proxy);
          }
        }
      }

      proxy = proxy_once.get_data ();
      if (proxy == null)
      {
        if (ignore_proxy)
        {
          // retry next time
          proxy_once.reset ();
        }
        var msg = "Unable to create scope proxy for \"%s\": %s".printf (
          scope_id, failure != null ? failure.message : "");
        throw new ScopeError.REQUEST_FAILED (msg);
      }

      // open a channel
      var channel_key = get_channel_key (master_channel_id, proxy);
      var channel_id_once = get_channel_id_once (channel_key);

      if (!channel_id_once.is_initialized ())
      {
        if (yield channel_id_once.enter ())
        {
          Dee.SerializableModel model;
          string? chan_id = null;
          try
          {
            chan_id = yield proxy.open_channel (channel_type,
                                                ChannelFlags.PRIVATE,
                                                null,
                                                out model);
            scope_models[channel_key] = model;
            model.set_data<string> ("scope-id", scope_id);
            // register as receiver
            var synchronizer = synchronizers[master_channel_id];
            if (synchronizer != null)
            {
              var uid = "%p".printf (proxy);
              scope_proxy_uids[uid] = proxy;
              synchronizer.add_provider (model, uid);
            }
            else
            {
              warning ("Unable to find ResultsSynchronizer for channel %s",
                       master_channel_id);
            }
          }
          finally
          {
            channel_id_once.leave (chan_id);
          }
          // emit a signal
          scope_channel_opened (master_channel_id, proxy, model);
        }
      }

      unowned string scope_channel_id = channel_id_once.get_data ();
      if (scope_channel_id == null)
      {
        // uh oh, couldn't open a channel, try again next time
        channel_id_once.reset ();
      }
      return scope_channel_id;
    }

    public async HashTable<string, Variant> search_wrapper (
        string master_channel_id,
        ChannelType channel_type,
        string search_string,
        HashTable<string, Variant> hints,
        string scope_id,
        CategoryMerger category_merger) throws Error
    {
      ScopeProxy proxy;
      unowned string scope_channel_id;
      scope_channel_id = get_channel_id (master_channel_id, scope_id, out proxy);
      if (scope_channel_id == null)
        scope_channel_id = yield init_channel (master_channel_id, scope_id,
                                               channel_type, out proxy);

      var reply_hints = new HashTable<string, Variant> (str_hash, str_equal);

      if (!content_enabled (scope_has_remote[scope_id])
          || scope_channel_id == null)
      {
        return reply_hints;
      }

      if (category_merger is CategoryMergerByField)
      {
        (category_merger as CategoryMergerByField).map_subscope_categories (scope_id, proxy.categories_model);
      }

      var channel_key = get_channel_key (master_channel_id, proxy);
      var last_seq_num = scope_models[channel_key].get_seqnum ();
      var reply_dict = yield proxy.search (scope_channel_id, search_string,
                                           hints);

      var iter = HashTableIter<string, Variant> (reply_dict);
      unowned string key;
      unowned Variant variant;

      while (iter.next (out key, out variant))
      {
        if (key == "model-seqnum")
        {
          uint64 seqnum = variant.get_uint64 ();
          var model = scope_models[channel_key];
          if (model.get_seqnum () < seqnum)
            yield wait_for_seqnum (model as Dee.SharedModel, seqnum);

          // if the proxy was disconnected and its channels invalidated, this
          // model is no longer merged, check if that's the case
          if (scope_models[channel_key] != model)
            return reply_hints;

          if (seqnum == last_seq_num)
          {
            debug ("Model seqnum for channel key %s not changed, copying", channel_key);
            var synchronizer = get_synchronizer (master_channel_id);
            if (synchronizer != null)
              synchronizer.copy_model (model);
            else
              warning ("No synchronizer for master channel %s", master_channel_id);
          }
        }
        else
        {
          reply_hints[key] = variant; // pass up
        }
      }

      return reply_hints;
    }

    public async void push_wrapper (
        string parent_channel_id,
        ChannelType channel_type,
        string master_scope_id,
        string scope_id,
        Dee.SerializableModel results_model,
        string[] categories) throws Error
    {
      ScopeProxy proxy;
      unowned string scope_channel_id;
      // we're yielding before using the array, need to copy it explicitely
      string[] categories_copy = categories;
      scope_channel_id = get_channel_id (parent_channel_id, master_scope_id, out proxy);
      if (scope_channel_id == null)
        scope_channel_id = yield init_channel (parent_channel_id,
                                               master_scope_id,
                                               channel_type, out proxy);

      if (scope_channel_id == null)
      {
        return; // shouldn't be reached really
      }

      yield proxy.push_results (scope_channel_id, scope_id, results_model,
                                categories_copy, null);
    }

    public signal void scope_channel_opened (string master_channel_id,
                                             ScopeProxy scope_proxy,
                                             Dee.SerializableModel model);

    public unowned ScopeProxy get_proxy_for_uid (string? uid) throws ScopeError
    {
      if (uid == null)
        throw new ScopeError.REQUEST_FAILED ("Invalid scope UID");

      unowned ScopeProxy proxy = scope_proxy_uids[uid];

      if (proxy != null) return proxy;

      throw new ScopeError.REQUEST_FAILED ("Invalid scope UID");
    }
  }

  private abstract class CategoryMerger : Object
  {
    protected HashTable<string, int> category_map = new HashTable<string, int> (str_hash, str_equal);
    public abstract int remap (string scope_id, uint32 category_index);
  }

  private class CategoryMergerByScope : CategoryMerger
  {
    public void add_scope_mapping (Unity.AggregatorScope scope, string scope_id)
    {
      var idx = scope.category_index_for_scope_id (scope_id);
      if (idx >= 0)
        category_map[scope_id] = idx;
    }

    public override int remap (string scope_id, uint32 category_index)
    {
      if (category_map.contains (scope_id))
        return category_map[scope_id];
      warning ("No category mapping for %s", scope_id);
      return -1;
    }
  }

  private class CategoryMergerByField : CategoryMerger
  {
    private HashTable<string, HashTable<int, int>?> subscopes = new HashTable<string, HashTable<int, int>?> (str_hash, str_equal);
    private uint column_index;

    public CategoryMergerByField (Dee.SerializableModel master_categories_model,
                                  uint column_index)
    {
      this.column_index = column_index;
      // iterate over aggregator scope categories and create name -> index mapping
      int count = 0;
      var iter = master_categories_model.get_first_iter ();
      var end_iter = master_categories_model.get_last_iter ();
      while (iter != end_iter)
      {
        var name = master_categories_model.get_string (iter, column_index);
        if (category_map.contains (name))
          warning ("Duplicated category name: %s", name);
        else
          category_map[name] = count++;
        iter = master_categories_model.next (iter);
      }
    }

    /**
     * Creates mapping for given subscope and its categories to aggregator scope category indexes
     */
    public void map_subscope_categories (string scope_id, Dee.SerializableModel categories_model)
    {
      string[] categories = new string [0];

      var iter = categories_model.get_first_iter ();
      var end_iter = categories_model.get_last_iter ();
      while (iter != end_iter)
      {
        string cat_id = categories_model.get_string (iter, column_index);
        categories += cat_id;
        iter = categories_model.next (iter);
      }
      map_subscope_categories_from_list (scope_id, categories);
    }

    public void map_subscope_categories_from_list (string scope_id, string[] categories)
    {
      int count = 0;

      HashTable<int, int> subscope_category_map = subscopes.lookup (scope_id);
      if (subscope_category_map == null)
      {
        subscope_category_map = new HashTable<int, int> (direct_hash, direct_equal);
        subscopes[scope_id] = subscope_category_map;
      }

      foreach (var cat_id in categories)
      {
        if (category_map.contains (cat_id)) //does master have this category?
        {
          var cat_idx = category_map[cat_id];
          subscope_category_map[count++] = cat_idx;
        }
        else
        {
          warning ("Subscope '%s' category '%s' not present in aggregator scope model", scope_id, cat_id);
        }
      }
    }

    public override int remap (string scope_id, uint32 category_index)
    {
      var subscope_category_map = subscopes.lookup (scope_id);
      if (subscope_category_map != null)
      {
        if (subscope_category_map.contains ((int)category_index))
        {
          var idx = subscope_category_map[(int)category_index];
          return (int)idx;
        }
        warning ("No category mapping for %s, category %u", scope_id, category_index);
      }
      warning ("No category mapping for %s", scope_id);
      return -1;
    }
  }

  private struct Sorter
  {
    int category;
    int column_index;
    string? field_name;
    char schema;
    int multiplier; // +1 / -1 to implement ASCENDING / DESCENDING sort

    public int apply (Variant[] row1, Variant[] row2)
    {
      Variant dummy1;
      Variant dummy2;
      unowned Variant var1 = row1[column_index];
      unowned Variant var2 = row2[column_index];
      if (field_name != null)
      {
        dummy1 = var1.lookup_value ("content", VariantType.VARDICT).lookup_value (field_name, null);
        dummy2 = var2.lookup_value ("content", VariantType.VARDICT).lookup_value (field_name, null);
        // handle non-presence of optional fields
        if (dummy1 == null && dummy2 == null) return 0;
        if (dummy1 == null || dummy2 == null)
          return dummy1 == null ? 1 : -1;
        var1 = dummy1;
        var2 = dummy2;
      }
      int result;
      switch (schema)
      {
        case 's':
          // FIXME: implement a way to use simple ascii sort
          result = var1.get_string ().collate (var2.get_string ());
          break;
        case 'i':
          result = compare_int (var1.get_int32 (), var2.get_int32 ());
          break;
        case 'u':
          result = compare_uint (var1.get_uint32 (), var2.get_uint32 ());
          break;
        case 'x':
          result = compare_int64 (var1.get_int64 (), var2.get_int64 ());
          break;
        case 't':
          result = compare_uint64 (var1.get_uint64 (), var2.get_uint64 ());
          break;
        case 'd':
          result = compare_double (var1.get_double (), var2.get_double ());
          break;
        default:
          return 0;
      }

      return result * multiplier;
    }

    private static int compare_int (int a, int b)
    {
      return a > b ? 1 : a == b ? 0 : -1;
    }

    private static int compare_uint (uint a, uint b)
    {
      return a > b ? 1 : a == b ? 0 : -1;
    }

    private static int compare_int64 (int64 a, int64 b)
    {
      return a > b ? 1 : a == b ? 0 : -1;
    }

    private static int compare_uint64 (uint64 a, uint64 b)
    {
      return a > b ? 1 : a == b ? 0 : -1;
    }

    private static int compare_double (double a, double b)
    {
      return a > b ? 1 : a == b ? 0 : -1;
    }
  }

  private const string SOURCES_FILTER_ID = "unity-sources";

  public unowned AggregatorScope owner { private get; construct; }
  private HashTable<string, ScopeChannel> _channels;
  private uint _dbus_id;
  private DBusConnection? _dbus_connection;
  private ScopeTracker _scopes;
  private Sorter[] _sorters;
  private Sorter[] _constraints;
  private CategoryMerger category_merger;
  private Rand _rand;

  public Dee.SerializableModel categories_model { get; set; }
  public Dee.SerializableModel filters_model { get; set; }

  /* Make sure the object doesn't reference itself, we do want it to be
   * properly destroyed */
  private MergeStrategy? _merge_strategy = null;
  public MergeStrategy? merge_strategy
  {
    internal get { return _merge_strategy != null ? _merge_strategy : this; }
    set { _merge_strategy = value != this ? value : null; }
  }

  /* we need notifications on this property */
  public ViewType view_type { get; set; }

  public AggregatorScopeImpl (AggregatorScope owner)
  {
    Object (owner: owner);
  }

  construct
  {
    _rand = new Rand ();
    _channels = new HashTable<string, ScopeChannel> (str_hash, str_equal);
    _scopes = new ScopeTracker ();
    merge_strategy = this;
    create_models ();
  }

  // keep the scope list in a set for fast lookup
  private static Gee.Set<string> disabled_scope_ids;

  private static bool is_scope_disabled (string scope_id)
  {
    if (disabled_scope_ids == null)
    {
      disabled_scope_ids = new Gee.HashSet<string> ();
      var pref_man = PreferencesManager.get_default ();
      pref_man.notify["disabled-scopes"].connect (update_disabled_scopes);
      update_disabled_scopes ();
    }

    return scope_id in disabled_scope_ids;
  }

  private static void update_disabled_scopes ()
  {
    var pref_man = PreferencesManager.get_default ();
    disabled_scope_ids.clear ();
    foreach (unowned string scope_id in pref_man.disabled_scopes)
    {
      disabled_scope_ids.add (scope_id);
    }
  }

  public GLib.List<weak string> subscope_ids ()
  {
    return _scopes.scope_ids_for_proxies ();
  }

  /* Create usable name prefix for the models */
  private string create_dbus_name ()
  {
    /* We randomize the names to avoid conflicts to ensure that we always
     * have a clean start (no processes hanging around that would cause
     * our models to not be the leaders)
     */
    var t = get_monotonic_time ();
    const string format_string = "com.canonical.Unity.Master.Scope.%s.T%" + int64.FORMAT + "%d";
    var dbus_name = format_string.printf (Path.get_basename (owner.dbus_path),
                                          t, _rand.int_range (0, 10000));
    return dbus_name;
  }

  private void create_models ()
  {
    /* Schema definitions come from the Lens specification */
    categories_model = new Dee.SequenceModel ();
    categories_model.set_schema_full (CATEGORIES_SCHEMA);

    filters_model = new Dee.SequenceModel ();
    filters_model.set_schema_full (FILTERS_SCHEMA);
  }

  public void export () throws IOError
  {
    _dbus_connection = Bus.get_sync (BusType.SESSION);
    _dbus_id = _dbus_connection.register_object (owner.dbus_path,
                                                 this as ScopeService);
  }

  public void unexport ()
  {
    if (_dbus_id != 0)
    {
      _dbus_connection.unregister_object (_dbus_id);
      _dbus_id = 0;
      _dbus_connection = null;
    }

    // get rid of open channels
    string[] channel_ids = {};
    foreach (unowned string channel_id in _channels.get_keys ())
    {
      channel_ids += channel_id;
    }
    foreach (unowned string channel_id in channel_ids)
    {
      _scopes.unregister_channel (channel_id);
      _channels.remove (channel_id);
    }
  }

  public void set_categories (List<Category> categories)
  {
    bool categories_model_empty = categories_model.get_n_rows () == 0;
    if (!categories_model_empty)
    {
      // we support only appending new categories, no changes/deletes
      unowned List<Category> cats = categories;
      uint cats_length = categories.length ();
      bool data_matches = cats_length >= categories_model.get_n_rows ();

      var iter = categories_model.get_first_iter ();
      var end_iter = categories_model.get_last_iter ();
      while (data_matches && iter != end_iter)
      {
        data_matches &= cats.data.id == categories_model.get_string (iter, 0);
        // FIXME: emit row-changed if other props changed
        iter = categories_model.next (iter);
        cats = cats.next;
      }

      if (!data_matches)
      {
        warning ("Categories can only be added, ignoring request");
        return;
      }
      else
      {
        categories = cats;
      }
    }

    foreach (unowned Category category in categories)
    {
      string icon_hint = Utils.icon_to_string (category.icon_hint);
      categories_model.append (category.id,
                               category.name,
                               icon_hint,
                               category.renderer,
                               Utils.hash_table_to_asv (category.hints));
    }

    queue_property_notification ("Categories",
                                 new Variant.variant (this.categories));
  }

  public void set_filters (List<Filter> filters)
  {
    filters_model.clear ();

    foreach (unowned Filter filter in filters)
    {
      //filter.changed.connect (on_filter_option_changed);
    }

    List<unowned Filter> filters_and_sources = filters.copy ();
    if (owner.sources.options.length () > 0)
      filters_and_sources.append (owner.sources);

    Variant data[8];
    foreach (unowned Filter filter in filters_and_sources)
    {
      var serialized_filter = filter.serialize ();
      for (size_t i = 0; i < serialized_filter.n_children (); i++)
        data[i] = serialized_filter.get_child_value (i);

      filters_model.append_row (data);
    }
  }

  public void queue_search_for_type (SearchType search_type)
  {
    // not really relevant for master scopes
  }

  public void invalidate_search (SearchType search_type)
  {
    // not really relevant for master scopes
  }

  private VariantBuilder? changed_props;

  public void queue_property_notification (string prop_name, Variant prop_value)
  {
    if (_dbus_id == 0) return;

    bool schedule_emit = changed_props == null;
    if (changed_props == null)
      changed_props = new VariantBuilder (new VariantType ("a{sv}"));

    changed_props.add ("{sv}", prop_name, prop_value);

    if (schedule_emit)
    {
      Idle.add (() =>
      {
        var invalidated = new Variant.array (new VariantType ("s"), {});
        try
        {
          _dbus_connection.emit_signal (null, owner.dbus_path,
                                        "org.freedesktop.DBus.Properties",
                                        "PropertiesChanged",
                                        new Variant ("(sa{sv}@as)",
                                                     ScopeService.INTERFACE_NAME,
                                                     changed_props,
                                                     invalidated));
        }
        catch (Error err)
        {
          warning ("%s", err.message);
        }
        changed_props = null;
        return false;
      });
    }
  }



  public void add_sorter (uint category_index, string field,
                          AggregatorScope.SortFlags flags)
  {
    string? schema = null;
    bool is_base_column = field in RESULTS_COLUMN_NAMES;
    bool field_found = false;

    foreach (unowned Schema.FieldInfo? info in owner.schema.get_fields ())
    {
      if (info.name == field)
      {
        schema = info.schema;
        field_found = true;
        break;
      }
    }
    // ensure the field name is valid
    if (!is_base_column && !field_found)
    {
      critical ("Field name '%s' is not valid for this scope", field);
      return;
    }

    // get column index
    int col_index = -1;
    for (int i = 0; i < RESULTS_COLUMN_NAMES.length; i++)
    {
      if (field == RESULTS_COLUMN_NAMES[i])
      {
        col_index = i;
        break;
      }
    }

    if (col_index >= 0)
      schema = RESULTS_SCHEMA[col_index];

    if (new VariantType (schema).is_basic () == false)
    {
      critical ("Only basic types can be sorted, '%s' is not supported",
                schema);
      return;
    }

    var sorter = Sorter ();
    sorter.category = (int) category_index;
    sorter.column_index = col_index < 0 ? RESULTS_COLUMN_NAMES.length - 1 : col_index;
    sorter.field_name = col_index < 0 ? field : null;
    sorter.schema = schema[0];
    sorter.multiplier = AggregatorScope.SortFlags.DESCENDING in flags ? -1 : 1;

    _sorters += (owned) sorter;
  }

  public void add_constraint (int category_index, string field)
  {
    string? schema = null;
    bool is_base_column = field in RESULTS_COLUMN_NAMES;
    bool field_found = false;

    foreach (unowned Schema.FieldInfo? info in owner.schema.get_fields ())
    {
      if (info.name == field)
      {
        schema = info.schema;
        field_found = true;
        break;
      }
    }
    // ensure the field name is valid
    if (!is_base_column && !field_found)
    {
      critical ("Field name '%s' is not valid for this scope", field);
      return;
    }

    // get column index
    int col_index = -1;
    for (int i = 0; i < RESULTS_COLUMN_NAMES.length; i++)
    {
      if (field == RESULTS_COLUMN_NAMES[i])
      {
        col_index = i;
        break;
      }
    }

    if (col_index >= 0)
      schema = RESULTS_SCHEMA[col_index];

    if (new VariantType (schema).is_basic () == false)
    {
      critical ("Only basic types can be sorted, '%s' is not supported",
                schema);
      return;
    }

    if (category_index >= 0)
    {
      // we need a category sorter
      bool category_sorter_present = false;
      int category_col_index = -1;
      for (int i = 0; i < RESULTS_COLUMN_NAMES.length; i++)
      {
        if (RESULTS_COLUMN_NAMES[i] == "category")
        {
          category_col_index = i;
          break;
        }
      }
      warn_if_fail (category_col_index >= 0);
      foreach (unowned Sorter it_sorter in _constraints)
      {
        if (it_sorter.column_index == category_col_index)
        {
          category_sorter_present = true;
          break;
        }
      }
      if (!category_sorter_present) add_constraint (-1, "category");
    }

    var sorter = Sorter ();
    sorter.category = category_index;
    sorter.column_index = col_index < 0 ? RESULTS_COLUMN_NAMES.length - 1 : col_index;
    sorter.field_name = col_index < 0 ? field : null;
    sorter.schema = schema[0];
    sorter.multiplier = 1;

    foreach (unowned Sorter it_sorter in _constraints)
    {
      // check for duplicate constraints
      if (it_sorter.category == sorter.category && 
          it_sorter.column_index == sorter.column_index &&
          it_sorter.field_name == sorter.field_name &&
          it_sorter.schema == sorter.schema)
      {
        warning ("Trying to add duplicate constraint, ignoring");
        return;
      }
    }

    _constraints += (owned) sorter;
  }

  private static int apply_sorters (Variant[] row1, Variant[] row2,
                                    uint category, Sorter[] sorters)
  {
    foreach (unowned Sorter sorter in sorters)
    {
      if (sorter.category >= 0 && sorter.category != category) continue;
      int sorter_result = sorter.apply (row1, row2);
      if (sorter_result != 0) return sorter_result;
    }
    return 0;
  }

  /* MergeStrategy implementation */
  private unowned Dee.ModelIter? merge_result (string scope_id,
                                               Dee.Model target, Variant[] row)
  {
    // remap category index in the result
    uint row_category = row[ResultColumn.CATEGORY].get_uint32 ();
    int cat_idx = category_merger.remap (scope_id, row_category);
    if (cat_idx >= 0)
    {
      row[ResultColumn.CATEGORY] = new GLib.Variant.uint32 (cat_idx);
    }
    else
    {
      warning ("Unable to remap category %u for %s", row_category, scope_id);
      return null;
    }

    // check for duplicates
    if (_constraints.length > 0)
    {
      /* the lookups need to be fast, so keeping a FilterModel on top
       * of the real model that's sorted according to the de-dup fields */
      var dedup_model = target.get_qdata<Dee.Model> (dedup_model_quark);
      if (dedup_model == null)
      {
        warn_if_fail (target.get_n_rows () == 0);
        var filter = Dee.Filter.new (() => {}, // no MapFunc
                                     (orig_model, orig_iter, m) =>
        {
          Variant row_data[9];
          // FIXME: add a get_row_static method to dee
          var dummy = orig_model.get_row (orig_iter, row_data);
          void* dont_free = (owned) dummy;
          var iter = m.find_row_sorted (row_data, (row1, row2) =>
          {
            uint category = row1[ResultColumn.CATEGORY].get_uint32 ();
            return apply_sorters (row1, row2, category, _constraints);
          }, null);
          m.insert_iter_before (orig_iter, iter);
          return true;
        });
        dedup_model = new Dee.FilterModel (target, filter);
        // eek, circular reference
        target.set_qdata (dedup_model_quark, dedup_model);
      }
      bool found;
      bool fell_through = false;
      dedup_model.find_row_sorted (row, (row1, row2) =>
      {
        uint category = row1[ResultColumn.CATEGORY].get_uint32 ();
        // customized version of apply_sorters that's suitable for dedup
        for (int i = 0; i < _constraints.length; i++)
        {
          unowned Sorter sorter = _constraints[i];
          if (sorter.category >= 0 && sorter.category != category) continue;
          int sorter_result = sorter.apply (row1, row2);
          if (sorter_result != 0 || i == _constraints.length - 1)
            return sorter_result;
        }
        fell_through = true;
        return 0;
      }, out found);
      if (found && !fell_through) return null;
    }

    return target.insert_row_sorted (row, (row1, row2) =>
    {
      // first of all sort by categories
      uint category = row1[ResultColumn.CATEGORY].get_uint32 ();
      int cat_delta = (int) category -
                      (int) row2[ResultColumn.CATEGORY].get_uint32 ();
      if (cat_delta != 0) return cat_delta > 0 ? 1 : -1;

      return apply_sorters (row1, row2, category, _sorters);
    });
  }

  private unowned ScopeProxy? get_proxy_for_result (Variant? result_metadata,
                                                    out string? scope_id)
    throws ScopeError
  {
    if (result_metadata == null)
      throw new ScopeError.REQUEST_FAILED ("Unable to find scope proxy");

    string scope_uid;
    if (!result_metadata.lookup ("scope-uid", "s", out scope_uid))
      throw new ScopeError.REQUEST_FAILED ("Unable to find scope proxy");

    if (scope_uid.has_prefix ("0x"))
    {
      unowned ScopeProxy result = _scopes.get_proxy_for_uid (scope_uid);
      scope_id = _scopes.scope_id_for_proxy (result);
      return result;
    }
    else if (scope_uid.has_prefix ("remote:"))
    {
      // special case the proxy for results that were pushed
      var parts = scope_uid.split (":", 2);
      scope_id = parts[1];
      return null;
    }

    throw new ScopeError.REQUEST_FAILED ("Invalid scope UID");
  }


  /*
   * DBus Interface Implementation
   */
  public async ActivationReplyRaw activate (
      string channel_id,
      Variant[] result_arr,
      uint action_type,
      HashTable<string, Variant> hints,
      GLib.Cancellable? cancellable) throws IOError, ScopeError
  {
    get_channel_by_id (channel_id); // ensure the channel is valid

    string? scope_id;
    unowned ScopeProxy proxy;
    Variant metadata = result_arr[ResultColumn.METADATA];
    proxy = get_proxy_for_result (metadata, out scope_id);
    result_arr[ResultColumn.METADATA] = metadata.lookup_value ("content", null);

    var scope_result = ScopeResult ();
    scope_result.uri = result_arr[ResultColumn.URI].get_string ();
    scope_result.icon_hint = result_arr[ResultColumn.ICON_HINT].get_string ();
    scope_result.category = result_arr[ResultColumn.CATEGORY].get_uint32 ();
    scope_result.result_type = (ResultType) result_arr[ResultColumn.RESULT_TYPE].get_uint32 ();
    scope_result.mimetype = result_arr[ResultColumn.MIMETYPE].get_string ();
    scope_result.title = result_arr[ResultColumn.TITLE].get_string ();
    scope_result.comment = result_arr[ResultColumn.COMMENT].get_string ();
    scope_result.dnd_uri = result_arr[ResultColumn.DND_URI].get_string ();
    if (result_arr[ResultColumn.METADATA].get_type ().equal (VariantType.VARDICT))
    {
      scope_result.metadata = (HashTable<string, Variant>) result_arr[ResultColumn.METADATA];
    }

    if (scope_id == null)
    {
      throw new ScopeError.REQUEST_FAILED ("Unable to determine scope id!");
    }

    // arrays in async methods are not copied by default, force the copy
    var result_arr_cpy = result_arr;
    var activation_obj = new AggregatorActivation (channel_id, scope_id,
                                                   action_type, scope_result);

    var response = yield owner.activate (activation_obj);
    if (response != null)
    {
      var raw = ActivationReplyRaw ();
      raw.uri = scope_result.uri;
      raw.handled = response.handled;
      raw.hints = response.get_hints ();

      return raw;
    }

    if (proxy == null)
    {
      if (action_type != Protocol.ActionType.PREVIEW_RESULT)
      {
        response = new ActivationResponse (HandledType.NOT_HANDLED);
      }
      else
      {
        response = new ActivationResponse.with_preview (GenericPreview.empty ());
      }
      var raw = ActivationReplyRaw ();
      raw.uri = scope_result.uri;
      raw.handled = response.handled;
      raw.hints = response.get_hints ();

      return raw;
    }

    /* proxy the request by default */
    var child_channel = _scopes.get_child_scope_channel_id (channel_id, proxy);
    try
    {
      var action = (Unity.Protocol.ActionType) action_type;
      return yield proxy.activate (child_channel, result_arr_cpy,
                                   action, hints, cancellable);
    }
    catch (Error err)
    {
      throw new ScopeError.REQUEST_FAILED ("Unhandled error: %s".printf (err.message));
    }
  }

  public async HashTable<string, Variant> search (
      string channel_id,
      string search_string,
      HashTable<string, Variant> hints,
      GLib.Cancellable? cancellable) throws IOError, ScopeError
  {
    var channel = get_channel_by_id (channel_id);

    // clear the master/aggregated model so that results from scopes that are not searched this time disappear
    var synchronizer = _scopes.get_synchronizer (channel_id);
    if (synchronizer != null)
      synchronizer.clear ();
    else
      warning ("No synchronizer for channel %s", channel_id);

    unowned Variant filter_row_variant = hints["changed-filter-row"];
    if (filter_row_variant != null)
      update_filter_state (channel, filter_row_variant);

    var response = new HashTable<string, Variant> (str_hash, str_equal);

    SearchContext search_context = SearchContext ();
    search_context.search_query = search_string;
    search_context.search_type = channel.get_search_type ();
    search_context.filter_state = channel.filters;
    search_context.search_metadata = SearchMetadata.create (hints);
    search_context.result_set = new DeeResultSet.with_model (channel.results_model);
    search_context.result_set.ttl = -1;
    search_context.cancellable = Unity.Cancellable.create ();

    var aggsearch = new AggregatedScopeSearch (owner, channel.id,
                                               hints, channel.results_model);
    aggsearch.set_search_context (search_context);

    ulong sig_id = aggsearch.category_order_changed.connect ((indices) =>
    {
      if (indices.length != categories_model.get_n_rows ())
        warning ("Invalid number of category indices, expected %u, got %u", categories_model.get_n_rows (), indices.length);
      else
        category_order_changed (channel_id, indices);
    });

    channel.last_search = aggsearch;

    aggsearch.run_async (() =>
    {
      Idle.add_full (Priority.DEFAULT, search.callback);
    });
    yield; // call virtual search method of the public API

    SignalHandler.disconnect (aggsearch, sig_id);

    // copy reply hints into response hash
    HashTable<string, Variant>? reply_hints = aggsearch.get_reply_hints ();
    if (reply_hints != null)
    {
      HashTableIter<string, Variant> iter = HashTableIter<string, Variant> (reply_hints);
      unowned string key;
      unowned Variant variant;
      while (iter.next (out key, out variant))
        response.insert (key, variant);
    }
    response["model-seqnum"] = new Variant.uint64 (channel.results_model.get_seqnum ());

    if (cancellable != null) cancellable.set_error_if_cancelled ();
    return response;
  }

  public async HashTable<string, Variant> search_scope (
      Unity.AggregatedScopeSearch search, string scope_id,
      string search_string, SearchType search_type,
      HashTable<string, Variant>? hints) throws Error
  {
    if (is_scope_disabled (scope_id))
    {
      return new HashTable<string, Variant> (null, null);
    }

    // remove the filter hints
    if (!owner.proxy_filter_hints)
      search.hints.remove ("changed-filter-row");

    HashTable<string, Variant>? combined_hints = null;
    if (hints != null)
    {
      combined_hints = new HashTable<string, Variant> (str_hash, str_equal);
      HFunc<string, Variant> combiner = (k, v) =>
      {
        combined_hints.insert (k, v);
      };
      hints.foreach (combiner);
      search.hints.foreach (combiner);
    }

    if (category_merger is CategoryMergerByScope)
      (category_merger as CategoryMergerByScope).add_scope_mapping (owner, scope_id);

    var channel_type = search_type == SearchType.DEFAULT ?
      ChannelType.DEFAULT : ChannelType.GLOBAL;
    var res = yield _scopes.search_wrapper (search.channel_id,
                                            channel_type,
                                            search_string,
                                            combined_hints != null ? combined_hints : search.hints, scope_id,
                                            category_merger);
    return res;
  }

  public async void push_results_to_scope (string channel_id,
                                           string scope_id,
                                           Dee.SerializableModel results_model,
                                           string[] category_ids) throws Error
  {
    if (!("-" in scope_id))
    {
      throw new ScopeError.REQUEST_FAILED ("Incorrect scope_id \"%s\"". printf (scope_id));
    }

    if (is_scope_disabled (scope_id))
    {
      throw new ScopeError.REQUEST_FAILED ("Scope is disabled");
    }

    // get the master scope id
    var parts = scope_id.split ("-", 2);
    var master_scope_id = "%s.scope".printf (parts[0]);
    
    if (category_merger is CategoryMergerByScope)
      (category_merger as CategoryMergerByScope).add_scope_mapping (owner, master_scope_id);

    yield _scopes.push_wrapper (channel_id, ChannelType.GLOBAL,
                                master_scope_id, scope_id,
                                results_model, category_ids);
  }

  public void push_filter_settings (string channel_id, FilterSet filters)
  {
    try
    {
      get_channel_by_id (channel_id);
    }
    catch (Error err)
    {
      warning ("Unexpected error: %s", err.message);
    }

    var filter_rows = new VariantBuilder (new VariantType ("a(ssssa{sv}bbb)"));
    foreach (unowned Filter f in filters.get_filters ())
    {
      filter_rows.add_value (f.serialize ());
    }

    filter_settings_changed (channel_id, filter_rows.end ());
  }

  private void update_filter_state (ScopeChannel channel,
                                    Variant changed_row) throws ScopeError
  {
    if (changed_row.get_type_string () != "(ssssa{sv}bbb)")
    {
      throw new ScopeError.REQUEST_FAILED ("Incorrect signature of filter-state (got '%s')".printf (changed_row.get_type_string ()));
    }

    string filter_id;
    changed_row.get_child (FilterColumn.ID, "s", out filter_id);
    var filter = channel.get_filter_by_id (filter_id);

    if (filter == null)
    {
      throw new ScopeError.REQUEST_FAILED ("Unable to find filter with id '%s'".printf (filter_id));
    }
    // update() will just update the hints, need to handle base props

    bool state;
    changed_row.get_child (FilterColumn.FILTERING, "b", out state);
    filter.filtering = state;
    changed_row.get_child (FilterColumn.COLLAPSED, "b", out state);
    filter.collapsed = state;
    filter.update (changed_row.get_child_value (FilterColumn.RENDERER_STATE));
  }

  public async string open_channel (
      uint channel_type,
      HashTable<string, Variant> hints,
      GLib.Cancellable? cancellable,
      out HashTable<string, Variant> out_hints) throws IOError
  {
    Unity.AggregatorScope aggscope = owner as Unity.AggregatorScope;

    /* Create category merger */
    if (aggscope.merge_mode == AggregatorScope.MergeMode.OWNER_SCOPE)
    {
      category_merger = new CategoryMergerByScope ();
    }
    else
    {
      category_merger = new CategoryMergerByField (categories_model, CategoryColumn.ID);
    }

    ChannelFlags flags = "private-channel" in hints ?
      ChannelFlags.PRIVATE : 0;
    var channel = new ScopeChannel ((ChannelType) channel_type);

    var schema = owner.schema;
    var required_schema = new HashTable<string, string> (str_hash, str_equal);
    var optional_schema = new HashTable<string, string> (str_hash, str_equal);
    foreach (unowned Unity.Schema.FieldInfo? field in schema.get_fields ())
    {
      if (field.type == Schema.FieldType.REQUIRED)
        required_schema[field.name] = field.schema;
      else
        optional_schema[field.name] = field.schema;
    }

    var model_name = channel.create_channel (create_dbus_name (),
                                             required_schema,
                                             optional_schema,
                                             filters_model,
                                             flags | ChannelFlags.NO_FILTERING);
    var sm = channel.results_model as Dee.SharedModel;
    if (sm != null)
    {
      // force AUTOMATIC flushing for the deprecated scopes, although we can
      // flush manually in the future
      sm.flush_mode = Dee.SharedModelFlushMode.AUTOMATIC;
    }

    _channels[channel.id] = channel;
    _scopes.register_channel (channel.id, channel.results_model, merge_strategy);

    out_hints = new HashTable<string, Variant> (str_hash, str_equal);
    out_hints["model-swarm-name"] = new Variant.string (model_name);

    return channel.id;
  }

  private ScopeChannel get_channel_by_id (string channel_id)
    throws ScopeError
  {
    unowned ScopeChannel channel = _channels[channel_id];
    if (channel == null)
      throw new ScopeError.INVALID_CHANNEL ("Invalid channel ID!");

    return channel;
  }

  public async void close_channel (
      string channel_id,
      HashTable<string, Variant> hints,
      GLib.Cancellable? cancellable) throws IOError, ScopeError
  {
    unowned ScopeChannel channel = _channels[channel_id];
    if (channel == null)
      throw new ScopeError.INVALID_CHANNEL ("Invalid channel ID!");

    _scopes.unregister_channel (channel_id);
    _channels.remove (channel_id);
  }

  public async void set_view_type (uint view_type_id) throws IOError
  {
    ViewType view_type = (ViewType) view_type_id;
    this.view_type = view_type;
  }

  public async void set_active_sources (
      string channel_id, string[] sources,
      GLib.Cancellable? cancellable) throws IOError
  {
    owner.set_active_sources_internal (sources);
  }

  public async void push_results (
          string channel_id,
          string source_scope_id,
          Variant model_v,
          string[] categories,
          GLib.Cancellable? cancellable = null) throws IOError, ScopeError
  {
    var model_obj = Dee.Serializable.parse (model_v, typeof (Dee.SequenceModel));
    if (model_obj == null)
      throw new ScopeError.REQUEST_FAILED ("Can't deserialize model");

    var provider = model_obj as Dee.SerializableModel;
    provider.set_data<string> ("uid", "remote:%s".printf (source_scope_id));
    provider.set_data<string> ("scope-id", source_scope_id);
    var sync = _scopes.get_synchronizer (channel_id);
    if (sync == null)
      throw new ScopeError.REQUEST_FAILED ("No synchronizer for channel %s", channel_id);

    var merger = category_merger as CategoryMergerByField;
    if (merger != null)
    {
      merger.map_subscope_categories_from_list (source_scope_id, categories);
    }
    else
    {
      throw new ScopeError.REQUEST_FAILED ("Merging not implemented");
    }
    
    var iter = provider.get_first_iter ();
    var end = provider.get_last_iter ();
    /* everything is prepared now, we can push results directly
     * into the synchronizer */
    while (iter != end)
    {
      unowned Variant[] row_buf = sync.prepare_row_buf (provider, iter);
      sync.merge_strategy.merge_result (source_scope_id, sync.receiver, row_buf);

      iter = provider.next (iter);
    }
    
    if (sync.receiver is Dee.SharedModel)
      (sync.receiver as Dee.SharedModel).flush_revision_queue ();
  }
 
  /* DBus properties */
  public bool visible { owned get { return owner.visible; } }
  public bool is_master { owned get { return owner.is_master; } }
  public string search_hint { owned get { return owner.search_hint ?? ""; } }
  public HashTable<string, string> metadata
  {
    owned get
    {
      var schema = owner.schema;
      var required_schema = new HashTable<string, string> (str_hash, str_equal);
      foreach (unowned Unity.Schema.FieldInfo? field in schema.get_fields ())
      {
        if (field.type == Schema.FieldType.REQUIRED)
          required_schema[field.name] = field.schema;
      }
      return required_schema;
    }
  }
  public HashTable<string, string> optional_metadata
  {
    owned get
    {
      var schema = owner.schema;
      var optional_schema = new HashTable<string, string> (str_hash, str_equal);
      foreach (unowned Unity.Schema.FieldInfo? field in schema.get_fields ())
      {
        if (field.type == Schema.FieldType.OPTIONAL)
          optional_schema[field.name] = field.schema;
      }
      return optional_schema;
    }
  }
  public Variant categories
  {
    owned get { return categories_model.serialize (); }
  }
  public Variant filters
  {
    owned get { return filters_model.serialize (); }
  }
  public HashTable<string, Variant> hints
  {
    owned get { return new HashTable<string, Variant> (null, null); }
  }
}

} /* namespace Unity.Internal */
