/*
 * Copyright (C) 2011 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 *
 */

using GLib;

namespace Unity {

public class ScopeDBusConnector : Object
{
  public unowned AbstractScope scope { get; construct; }

  public ScopeDBusConnector (AbstractScope scope)
  {
    Object (scope: scope);
  }

  private Internal.ScopeDBusImpl pimpl;
  private bool exported;
  private uint bus_name_own_handle;

  public void export () throws Error
  {
    if (!exported) // && can_export ()
    {
      if (pimpl == null) pimpl = new Internal.DefaultScopeDBusImpl (scope);
      own_name (); // this will also export the object
      exported = true;
    }
  }

  public void unexport ()
  {
    if (exported)
    {
      pimpl.unexport ();
      exported = false;
      unown_name ();
    }
  }

  private void own_name () throws Error
  {
    var dbus_name = scope.get_group_name ();
    // check if the name is owned already
    // FIXME: but track which names does this process own, cause trying to own
    // an already owned name will fail (even though the caller is the owner)!
    bool has_owner;
    DBusConnection bus = Bus.get_sync (BusType.SESSION);
    Variant result = bus.call_sync ("org.freedesktop.DBus",
                                    "/org/freedesktop/dbus",
                                    "org.freedesktop.DBus",
                                    "NameHasOwner",
                                    new Variant ("(s)", dbus_name),
                                    new VariantType ("(b)"),
                                    DBusCallFlags.NO_AUTO_START,
                                    -1);
    result.get ("(b)", out has_owner);
    if (has_owner)
    {
      throw new IOError.FAILED ("Name '%s' is already owned".printf (dbus_name));
    }

    bool got_name = false;
    var ml = new MainLoop ();
    bus_name_own_handle = Bus.own_name (
      BusType.SESSION, dbus_name, BusNameOwnerFlags.NONE, () =>
      {
        // connection acquired callback
        try
        {
          pimpl.export ();
        }
        catch (Error err)
        {
          warning ("Unable to export object: %s", err.message);
        }
      },
      () => { got_name = true; ml.quit (); },
      () => { ml.quit (); }
    );
    ml.run ();
    if (!got_name)
    {
      throw new IOError.FAILED ("Unable to own DBus name '%s'".printf (dbus_name));
    }
  }

  private void unown_name ()
  {
    if (bus_name_own_handle != 0)
    {
      Bus.unown_name (bus_name_own_handle);
      bus_name_own_handle = 0;
    }
  }

  private static MainLoop primary_loop;

  public static void run ()
  {
    if (primary_loop == null) primary_loop = new MainLoop ();
    primary_loop.run ();
  }

  public static void quit ()
  {
    if (primary_loop != null) primary_loop.quit ();
  }
}

} /* namespace */
