/*
 * Copyright (C) 2011 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 *
 */

using GLib;
using Gee;

namespace Unity.Internal {

/*
 * Connects to the necessary signals to reflect changes in the providers on the
 * receiver Model
 */
internal class ResultsSynchronizer : GLib.Object, MergeStrategy
{
  public unowned Dee.Model receiver { get; construct; }
  public MergeStrategy merge_strategy { get; set; }

  private Gee.ArrayList<unowned Dee.Model> _providers;
  private Gee.HashMap<unowned Dee.Model, Dee.ModelTag<unowned Dee.ModelIter>> _provider_tags;
  private Gee.HashMap<unowned Dee.Model, Dee.ModelTag<uint>> _clear_seqnum_tags;
  private Variant[] row_buf = new Variant[9];
  private Quark scope_id_quark = Quark.from_string ("scope-id");
  private uint _clear_seq_num = 1;

  internal ResultsSynchronizer (Dee.Model receiver)
  {
    Object(receiver:receiver);
  }

  construct
  {
    _providers = new Gee.ArrayList<unowned Dee.Model> ();
    _provider_tags = new Gee.HashMap<unowned Dee.Model, unowned Dee.ModelTag> ();
    _clear_seqnum_tags = new Gee.HashMap<unowned Dee.Model, unowned Dee.ModelTag> ();
    merge_strategy = this;
  }

  public void clear ()
  {
    ++_clear_seq_num;
    receiver.clear ();
  }
 
  public unowned Dee.ModelIter? merge_result (string source_scope_id,
                                              Dee.Model target, Variant[] row)
  {
    return target.append_row (row);
  }

  public void add_provider (Dee.Model provider, string uid)
  {
    provider.set_data<string>("uid", uid);

    _providers.add (provider);
    _provider_tags[provider] = new Dee.ModelTag<unowned Dee.ModelIter> (provider);
    _clear_seqnum_tags[provider] = new Dee.ModelTag<uint> (provider);

    provider.row_added.connect (on_row_added);
    provider.row_removed.connect (on_row_removed);
    provider.row_changed.connect (on_row_changed);
  }

  internal void clear_provider_model (string uid)
  {
    foreach (var p in _providers)
    {
      if (p.get_data<string>("uid") == uid)
      {
        p.clear ();
        break;
      }
    }
  }

  internal unowned Variant[] prepare_row_buf (Dee.Model provider,
                                              Dee.ModelIter iter)
  {
    row_buf[0] = provider.get_value (iter, 0);
    row_buf[1] = provider.get_value (iter, 1);
    row_buf[2] = provider.get_value (iter, 2);
    row_buf[3] = provider.get_value (iter, 3);
    row_buf[4] = provider.get_value (iter, 4);
    row_buf[5] = provider.get_value (iter, 5);
    row_buf[6] = provider.get_value (iter, 6);
    row_buf[7] = provider.get_value (iter, 7);

    var scope_entry = new Variant.dict_entry ("scope-uid",
                                              new Variant.variant (provider.get_data<string> ("uid")));
    var content_entry = new Variant.dict_entry ("content",
                                                new Variant.variant (provider.get_value (iter, 8)));

    row_buf[8] = new Variant.array (VariantType.VARDICT.element (),
                                    {scope_entry, content_entry});

    return row_buf;
  }

  internal void remove_provider (string uid)
  {
    foreach (var p in _providers)
    {
      if (p.get_data<string>("uid") == uid)
      {
          _provider_tags.unset (p);
          _clear_seqnum_tags.unset (p);
          _providers.remove (p);
          break;
      }
    }
  }

  private void on_row_added (Dee.Model provider, Dee.ModelIter iter)
  {
    unowned Dee.ModelIter? i;

    prepare_row_buf (provider, iter);
    i = merge_strategy.merge_result (provider.get_qdata<string>(scope_id_quark),
                                     receiver, row_buf);

    if (i != null)
    {
      var tag = _provider_tags[provider];
      tag [provider, iter] = i;

      var rem_tag = _clear_seqnum_tags[provider];
      rem_tag [provider, iter] = _clear_seq_num;
    }
  }

  private void on_row_removed (Dee.Model provider, Dee.ModelIter iter)
  {
    var rem_tag = _clear_seqnum_tags[provider];
    var tag = _provider_tags[provider];
    uint seq_num = rem_tag[provider, iter];
    if (seq_num == _clear_seq_num)
    {
      rem_tag [provider, iter] = 0;
      var riter = tag[provider, iter];

      if (riter != null)
        receiver.remove (riter);
    }
  }

  private void on_row_changed (Dee.Model provider, Dee.ModelIter iter)
  {
    var tag = _provider_tags[provider];
    var riter = tag[provider, iter];

    if (riter != null)
    {
      prepare_row_buf (provider, iter);
      receiver.set_row (riter, row_buf);
    }
    else
    {
      warning (@"Could not find row to change: $(provider.get_string (iter, 0))");
    }
  }

  internal void copy_model (Dee.Model provider)
  {
    if (!_providers.contains (provider))
      critical ("Requested unknown provider model");
    else
    {
      var end = provider.get_last_iter ();
      var iter = provider.get_first_iter ();
      while (iter != end)
      {
        on_row_added (provider, iter);
        iter = provider.next (iter);
      }
    }
  }
}


} /* namespace Unity */
