/*
 * Copyright (C) 2011 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 *
 */

using GLib;
using Gee;

namespace Unity.Internal {

/* Cateories: [id, display_name, icon, renderer_name, hints] */
internal const string[] CATEGORIES_SCHEMA = {"s", "s", "s", "s", "a{sv}"};

internal enum CategoryColumn
{
  ID,
  DISPLAY_NAME,
  ICON_HINT,
  RENDERER_NAME,
  HINTS,

  N_COLUMNS
}

/* Filters: [id, display_name, icon, renderer_name, hints,
             visible, collapsed, filtering] */
internal const string[] FILTERS_SCHEMA = {"s", "s", "s", "s", "a{sv}",
                                          "b", "b", "b"};

internal enum FilterColumn
{
  ID,
  DISPLAY_NAME,
  ICON_HINT,
  RENDERER_NAME,
  RENDERER_STATE,
  VISIBLE,
  COLLAPSED,
  FILTERING,

  N_COLUMNS
}

/* Results: [uri, icon, category, result_type, mimetype,
             title, comment, dnd_uri, metadata] */
internal const string[] RESULTS_SCHEMA = {"s", "s", "u", "u", "s",
                                          "s", "s", "s", "a{sv}"};
internal const string[] RESULTS_COLUMN_NAMES = {
  "uri", "icon_hint", "category", "result_type", "mimetype",
  "title", "comment", "dnd_uri",  "metadata"};

internal enum ResultColumn
{
  URI,
  ICON_HINT,
  CATEGORY,
  RESULT_TYPE,
  MIMETYPE,
  TITLE,
  COMMENT,
  DND_URI,
  METADATA,

  N_COLUMNS
}

namespace Utils
{
  private struct DelegateWrapper
  {
    public unowned SourceFunc callback;

    public DelegateWrapper (SourceFunc cb)
    {
      callback = cb;
    }
  }

  internal class AsyncOnce<G>
  {
    private enum OperationState
    {
      NOT_STARTED,
      IN_PROGRESS,
      DONE
    }

    private G inner;

    private OperationState state;
    private DelegateWrapper[] callbacks = {};

    public AsyncOnce ()
    {
      state = OperationState.NOT_STARTED;
    }

    public unowned G get_data ()
    {
      return inner;
    }

    public bool is_initialized ()
    {
      return state == OperationState.DONE;
    }

    public async bool enter ()
    {
      if (state == OperationState.NOT_STARTED)
      {
        state = OperationState.IN_PROGRESS;
        return true;
      }
      else if (state == OperationState.IN_PROGRESS)
      {
        yield wait_async ();
      }

      return false;
    }

    public void leave (G result)
    {
      if (state != OperationState.IN_PROGRESS)
      {
        warning ("Incorrect usage of AsyncOnce");
        return;
      }
      state = OperationState.DONE;
      inner = result;
      notify_all ();
    }

    /* Once probably shouldn't have this, but it's useful */
    public void reset ()
    {
      state = OperationState.NOT_STARTED;
      inner = null;
    }

    private void notify_all ()
    {
      foreach (unowned DelegateWrapper wrapper in callbacks)
      {
        wrapper.callback ();
      }
      callbacks = {};
    }

    private async void wait_async ()
    {
      callbacks += DelegateWrapper (wait_async.callback);
      yield;
    }
  }

  internal class AsyncMutex
  {
    private bool is_locked;
    private GLib.Queue<DelegateWrapper?> callbacks;

    public AsyncMutex ()
    {
      callbacks = new GLib.Queue<DelegateWrapper?> ();
      is_locked = false;
    }

    public bool try_lock ()
    {
      if (!is_locked)
      {
        is_locked = true;
        return true;
      }
      return false;
    }

    public async void lock ()
    {
      if (!is_locked)
      {
        is_locked = true;
        return;
      }
      yield wait_async ();
      is_locked = true;
    }

    public void unlock ()
    {
      if (!is_locked)
      {
        warning ("Unlock failed: AsyncMutex was already unlocked");
        return;
      }
      is_locked = false;
      notify ();
    }

    private void notify ()
    {
      if (callbacks.is_empty ()) return;
      DelegateWrapper? cb = callbacks.pop_head ();
      cb.callback ();
    }

    private async void wait_async ()
    {
      callbacks.push_tail (DelegateWrapper (wait_async.callback));
      yield;
    }
  }

  internal static Variant hash_table_to_asv (HashTable<string, Variant> hash)
  {
    var b = new VariantBuilder (new VariantType ("a{sv}"));

    var iter = HashTableIter<string, Variant> (hash);
    
    string key;
    Variant val;
    while (iter.next (out key, out val))
    {
      b.add ("{sv}", key, val);
    }

    return b.end ();
  }

  internal static string icon_to_string (Icon? icon)
  {
    return icon != null ? icon.to_string () : "";
  }
}

} /* namespace Unity.Internal */
