/*
 * Copyright (C) 2011-2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Michal Hruby <michal.hruby@canonical.com>
 *
 */

using Unity.Protocol;
using Unity.Test;

/* A bit of magic to get proper-ish fixture support */
public interface Fixture : Object
{
  class DelegateWrapper
  {
    TestDataFunc func;
    public DelegateWrapper (owned TestDataFunc f) { func = (owned) f; }
  }

  public virtual void setup () {}
  public virtual void teardown () {}

  [CCode (has_target = false)]
  public delegate void Callback<T> (T ptr);

  private static List<DelegateWrapper> _tests;

  public static unowned TestDataFunc create<F> (Callback<void*> cb)
    requires (typeof (F).is_a (typeof (Fixture)))
  {
    TestDataFunc functor = () =>
    {
      var type = typeof (F);
      var instance = Object.new (type) as Fixture;
      instance.setup ();
      cb (instance);
      instance.teardown ();
    };
    unowned TestDataFunc copy = functor;
    _tests.append (new DelegateWrapper ((owned) functor));
    return copy;
  }
  public static unowned TestDataFunc create_static<F> (Callback<F> cb)
  {
    return create<F> ((Callback<void*>) cb);
  }
}

public class Main
{
  const string DBUS_NAME = "com.canonical.Unity.Scope.Test";
  const string DBUS_PATH = "/com/canonical/Unity/Scope/Test";

  const string MASTER_SCOPE_DBUS_PATH = "/com/canonical/Unity/MasterScope/Test";
  const string HOME_SCOPE_DBUS_PATH = "/com/canonical/Unity/HomeScope/Test";

  public static int main (string[] args)
  {
    string gsettings_schema_dir = Config.BUILDDIR+"/data";

    Environment.set_variable ("XDG_DATA_DIRS", Config.TESTDIR+"/data", true);
    Environment.set_variable ("XDG_DATA_HOME", Config.TESTDIR+"/data", true);
    Environment.set_variable ("LIBUNITY_LENS_DIRECTORY", Config.TESTDIR, true);
    Environment.set_variable ("GSETTINGS_SCHEMA_DIR", gsettings_schema_dir, true);
    Environment.set_variable ("GSETTINGS_BACKEND", "memory", true);
    try {
      Process.spawn_command_line_sync ("glib-compile-schemas " + gsettings_schema_dir);
    } catch (SpawnError e) {
      stderr.printf ("%s\n", e.message);
      return 1;
    }

    Test.init (ref args);

    Test.add_data_func ("/Unit/Scope/Export",
        Fixture.create<ScopeInitTester> (ScopeInitTester.test_scope_export));
    Test.add_data_func ("/Unit/Scope/ProxyConnection",
        Fixture.create<ScopeInitTester> (ScopeInitTester.test_scope_proxy));
    Test.add_data_func ("/Unit/Scope/OpenChannel",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_open_channel));
    Test.add_data_func ("/Unit/Scope/CloseChannel",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_open_close_channel));
    Test.add_data_func ("/Unit/Scope/Metadata",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_metadata));
    Test.add_data_func ("/Unit/Scope/Search",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_search));
    Test.add_data_func ("/Unit/Scope/Activation",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_activation));
    Test.add_data_func ("/Unit/Scope/Preview",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_preview));
    Test.add_data_func ("/Unit/Scope/PreviewAction",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_preview_action));
    Test.add_data_func ("/Unit/Scope/Filters",
        Fixture.create<ScopeTester> (ScopeTester.test_scope_filters));

    Test.add_data_func ("/Unit/MasterScope/OpenChannel",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_open_channel));
    Test.add_data_func ("/Unit/MasterScope/CloseChannel",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_open_close_channel));
    Test.add_data_func ("/Unit/MasterScope/Search",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_search));
    Test.add_data_func ("/Unit/MasterScope/Activation",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_activation));
    Test.add_data_func ("/Unit/MasterScope/Sorting/Ascending",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_sorting_asc));
    Test.add_data_func ("/Unit/MasterScope/Sorting/Descending",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_sorting_desc));
    Test.add_data_func ("/Unit/MasterScope/SortingOptionalFields/Ascending",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_sorting_optional_asc));
    Test.add_data_func ("/Unit/MasterScope/SortingOptionalFields/Descending",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_sorting_optional_desc));
    Test.add_data_func ("/Unit/MasterScope/Dedup/RequiredFields",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_deduplication));
    Test.add_data_func ("/Unit/MasterScope/Dedup/RequiredFields/PerCategory",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_deduplication_per_category));
    Test.add_data_func ("/Unit/MasterScope/Filters",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_master_filters));
    Test.add_data_func ("/Unit/MasterScope/SubscopesFilterHint",
        Fixture.create<MasterScopeTester> (MasterScopeTester.test_subscopes_filter_hint));
    Test.add_data_func ("/Unit/AggregatorScope/CategoryOrderSignal",
        Fixture.create<AggregatorScopeTester> (AggregatorScopeTester.test_scope_category_order_signal));

    Test.run ();

    return 0;
  }

  // this will auto-disconnect signals when it goes out of scope
  public class SignalWrapper
  {
    unowned Object obj;
    ulong sig_id;

    public SignalWrapper (Object o, ulong signal_id)
    {
      obj = o;
      sig_id = signal_id;
    }

    ~SignalWrapper ()
    {
      SignalHandler.disconnect (obj, sig_id);
    }
  }

  class TestSearcher: Unity.ScopeSearchBase
  {
    public unowned TestScope owner { get; set; }

    public TestSearcher (TestScope scope)
    {
      Object (owner: scope);
    }

    public override void run ()
    {
      var ml = new MainLoop ();
      // simulating async search
      owner.search (this);
      Idle.add (() => { ml.quit (); return false; });
      ml.run ();
    }
  }

  class TestPreviewer: Unity.ResultPreviewer
  {
    public unowned TestScope owner { get; set; }

    public TestPreviewer (TestScope scope)
    {
      Object (owner: scope);
    }

    public override Unity.AbstractPreview? run ()
    {
      var ml = new MainLoop ();
      // simulating async search
      var preview = owner.preview (this);
      Idle.add (() => { ml.quit (); return false; });
      ml.run ();

      return preview;
    }
  }

  class TestScope: Unity.AbstractScope
  {
    public string dbus_path { get; construct set; }

    public TestScope (string dbus_path)
    {
      Object (dbus_path: dbus_path);
    }

    public override Unity.ScopeSearchBase create_search_for_query (Unity.SearchContext ctx)
    {
      var searcher = new TestSearcher (this);
      assert (ctx.search_metadata.locale != null);
      searcher.set_search_context (ctx);

      return searcher;
    }

    public signal void search (Unity.ScopeSearchBase search_ctx);

    public signal Unity.AbstractPreview? preview (Unity.ResultPreviewer previewer);
    public signal Unity.ActivationResponse? activate_uri (string uri);
    public signal Unity.ActivationResponse? activate_action (Unity.ScopeResult result, string action_id);

    public override Unity.ResultPreviewer create_previewer (Unity.ScopeResult result, Unity.SearchMetadata metadata)
    {
      var previewer = new TestPreviewer (this);
      previewer.set_scope_result (result);
      previewer.set_search_metadata (metadata);
      assert (metadata.locale != null);

      return previewer;
    }

    public override Unity.ActivationResponse? activate (Unity.ScopeResult result, Unity.SearchMetadata metadata, string? action_id)
    {
      if (action_id == null)
        return activate_uri (result.uri);
      else
        return activate_action (result, action_id);
    }

    public override Unity.CategorySet get_categories ()
    {
      return new Unity.CategorySet ();
    }

    public override Unity.FilterSet get_filters ()
    {
      var filters = new Unity.FilterSet ();
      filters.add (ScopeTester.create_filter ());

      return filters;
    }

    public override Unity.Schema get_schema ()
    {
      var schema = new Unity.Schema ();
      schema.add_field ("required_string", "s", Unity.Schema.FieldType.REQUIRED);
      schema.add_field ("required_int", "i", Unity.Schema.FieldType.REQUIRED);
      schema.add_field ("optional_string", "s", Unity.Schema.FieldType.OPTIONAL);

      return schema;
    }

    public override string get_group_name ()
    {
      return DBUS_NAME;
    }

    public override string get_unique_name ()
    {
      return dbus_path;
    }

    public override string get_search_hint ()
    {
      return "Search hint";
    }

    private Unity.ScopeDBusConnector connector;
    public void export ()
    {
      connector = new Unity.ScopeDBusConnector (this);
      connector.export ();
    }

    public void unexport ()
    {
      connector.unexport ();
    }
  }

  class ScopeInitTester: Object, Fixture
  {
    private TestScope scope;

    public static ScopeProxy? acquire_test_proxy (string path = DBUS_PATH,
                                                  string name = DBUS_NAME)
    {
      var ml = new MainLoop ();
      ScopeProxy? proxy = null;
      ScopeProxy.new_from_dbus.begin (name, path, null, (obj, res) =>
      {
        try
        {
          proxy = ScopeProxy.new_from_dbus.end (res);
        }
        catch (Error e) {}
        ml.quit ();
      });
      assert (run_with_timeout (ml));
      return proxy;
    }

    private void setup ()
    {
    }

    private void teardown ()
    {
      if (scope != null) scope.unexport ();
    }

    public void test_scope_export ()
    {
      scope = new TestScope (DBUS_PATH);
      try
      {
        scope.export ();
      }
      catch (Error err) { assert_not_reached (); }
    }

    public void test_scope_proxy ()
    {
      scope = new TestScope (DBUS_PATH);
      scope.export ();

      var proxy = acquire_test_proxy ();
      assert (proxy != null);

      assert (proxy.search_hint == "Search hint");

      ensure_destruction ((owned) proxy);
    }
  }

  class ScopeTester: ScopeInitTester, Fixture
  {
    private TestScope scope;
    private ScopeProxy proxy;

    private void setup ()
    {
      base.setup ();

      scope = new TestScope (DBUS_PATH);
      
      try
      {
        scope.export ();
      }
      catch (Error err) { assert_not_reached (); }

      proxy = ScopeInitTester.acquire_test_proxy ();
      assert (proxy != null);
    }

    private void teardown ()
    {
      ensure_destruction ((owned) proxy);
      scope.unexport ();
      ensure_destruction ((owned) scope);

      base.teardown ();
    }

    public static Variant[] scope_result_to_variant (Unity.ScopeResult result)
    {
      var v = new Variant[9];
      v[0] = result.uri;
      v[1] = result.icon_hint;
      v[2] = result.category;
      v[3] = (uint) result.result_type;
      v[4] = result.mimetype;
      v[5] = result.title;
      v[6] = result.comment;
      v[7] = result.dnd_uri;
      v[8] = result.metadata;

      return v;
    }

    public static Unity.Filter create_filter ()
    {
      var check_option_filter = new Unity.CheckOptionFilter ("check-options",
                                                             "Check options",
                                                             null,
                                                             true);
      check_option_filter.add_option ("option1", "Option #1", null);
      check_option_filter.add_option ("option2", "Option #2", null);
      check_option_filter.add_option ("option3", "Option #3", null);
      return check_option_filter;
    }

    public static string open_channel (ScopeProxy proxy,
                                       ChannelType channel_type,
                                       out Dee.SerializableModel model,
                                       bool wait_for_synchronization = false)
    {
      string? channel_id = null;
      Dee.Model? real_model = null;
      SignalWrapper[] signals = {};
      var ml = new MainLoop ();
      /* Need to use PRIVATE channel, cause standard SharedModel won't
       * synchronize properly when trying to connect to the model
       * from the same process (/bus address) */
      proxy.open_channel.begin (channel_type,
                                ChannelFlags.PRIVATE,
                                null,
                                (obj, res) =>
      {
        channel_id = proxy.open_channel.end (res, out real_model);
        if (wait_for_synchronization && real_model is Dee.SharedModel)
        {
          var shared_model = real_model as Dee.SharedModel;
          signals += new SignalWrapper (shared_model, 
            shared_model.notify["synchronized"].connect (() =>
          {
            ml.quit ();
          }));
          if (shared_model.is_synchronized ()) ml.quit ();
        }
        else
        {
          ml.quit ();
        }
      });

      assert (run_with_timeout (ml));
      signals = {};
      model = real_model as Dee.SerializableModel;
      return channel_id;
    }

    public static void close_channel (ScopeProxy proxy, string channel_id)
    {
      var ml = new MainLoop ();
      proxy.close_channel.begin (channel_id, null,
                                 (obj, res) =>
      {
        proxy.close_channel.end (res);
        ml.quit ();
      });

      assert (run_with_timeout (ml));
    }

    public static HashTable<string, Variant> perform_search (
        ScopeProxy proxy, string channel_id, string query,
        HashTable<string, Variant>? hints = null,
        Dee.SerializableModel? model = null,
        TestScope? target_scope = null,
        Func<Unity.SearchContext?>? add_results_cb = null)
    {
      SignalWrapper[] signals = {};
      var ml = new MainLoop ();
      HashTable<string, Variant>? reply_dict = null;
      proxy.search.begin (channel_id, query,
                          hints ?? new HashTable<string, Variant> (null, null),
                          null,
                          (obj, res) =>
      {
        try
        {
          reply_dict = proxy.search.end (res);
        }
        catch (Error err) {}
        ml.quit ();
      });

      bool got_search_signal = false;
      if (target_scope != null)
      {
        signals += new SignalWrapper (
            target_scope, target_scope.search.connect ((search) =>
        {
          got_search_signal = true;
          if (add_results_cb != null) add_results_cb (search.search_context);
        }));
      }

      assert (run_with_timeout (ml));
      assert (reply_dict != null);
      if (target_scope != null) assert (got_search_signal);

      // wait for the model to synchronize
      var variant = reply_dict["model-seqnum"];
      if (variant != null && model != null)
      {
        uint64 desired_seqnum = variant.get_uint64 ();
        if (desired_seqnum > model.get_seqnum ())
        {
          ml = new MainLoop ();
          var shared_model = model as Dee.SharedModel;
          signals += new SignalWrapper (shared_model,
            shared_model.end_transaction.connect ((seqnum1, seqnum2) =>
          {
            if (seqnum2 >= desired_seqnum)
              Idle.add (() => { ml.quit (); return false; });
          }));
          run_with_timeout (ml);
        }
      }
      signals = {};
      return reply_dict;
    }

    public static HashTable<string, Variant> preview_result (
        ScopeProxy proxy, string channel_id, Unity.ScopeResult result,
        TestScope? target_scope,
        Unity.Preview? result_preview = null)
    {
      SignalWrapper[] signals = {};
      var ml = new MainLoop ();
      HashTable<string, Variant>? reply_dict = null;
      var result_arr = scope_result_to_variant (result);
      proxy.activate.begin (channel_id, result_arr,
                            Unity.Protocol.ActionType.PREVIEW_RESULT,
                            new HashTable<string, Variant> (null, null),
                            null,
                            (obj, res) =>
      {
        try
        {
          var reply = proxy.activate.end (res);
          reply_dict = reply.hints;
        }
        catch (Error err) { warning ("%s", err.message); }
        ml.quit ();
      });

      bool got_preview_signal = false;
      if (target_scope != null)
      {
        signals += new SignalWrapper (
            target_scope, target_scope.preview.connect ((search) =>
        {
          got_preview_signal = true;
          return result_preview;
        }));
      }

      assert (run_with_timeout (ml));
      assert (reply_dict != null);
      if (target_scope != null) assert (got_preview_signal);

      signals = {};
      return reply_dict;
    }

    public void test_scope_open_channel ()
    {
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);
    }

    public void test_scope_open_close_channel ()
    {
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);
      close_channel (proxy, channel_id);
    }

    public void test_scope_metadata ()
    {
      Dee.Model model;
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, out model, true);
      assert (channel_id != null);

      assert (model.get_column_names () != null);
      assert (model.get_field_schema ("required_string", null) == "s");
      assert (model.get_field_schema ("required_int", null) == "i");
      assert (model.get_field_schema ("optional_string", null) == "s");
    }

    public void test_scope_search ()
    {
      // we need channel first
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);

      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null,
                                                   null, scope, (search) =>
      {
        Unity.ScopeResult result = Unity.ScopeResult ();
        result.uri = "test:uri";
        result.icon_hint = "";
        result.category = 0;
        result.result_type = Unity.ResultType.DEFAULT;
        result.mimetype = "inode/folder";
        result.title = "Title";
        result.comment = "";
        result.dnd_uri = "test::uri";
        result.metadata = new HashTable<string, Variant> (str_hash, str_equal);
        result.metadata["required_int"] = 5;
        result.metadata["required_string"] = "foo";
        search.result_set.add_result (result);
      });

      assert (reply_dict != null);
    }

    public void test_scope_preview ()
    {
      // we need channel first
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);

      Unity.ScopeResult result = Unity.ScopeResult ();
      result.uri = "test:uri";
      result.icon_hint = "";
      result.category = 0;
      result.result_type = Unity.ResultType.DEFAULT;
      result.mimetype = "inode/folder";
      result.title = "Title";
      result.comment = "";
      result.dnd_uri = "test:uri";
      result.metadata = new HashTable<string, Variant> (str_hash, str_equal);
      result.metadata["required_int"] = 5;
      result.metadata["required_string"] = "foo";

      var preview = new Unity.GenericPreview ("test:uri", "subtitle", null);

      var reply_dict = ScopeTester.preview_result (proxy, channel_id, result,
                                                   scope, preview);
      assert (reply_dict != null);
      var preview_v = reply_dict["preview"];
      assert (preview_v != null);

      var reconstructed = Unity.Protocol.Preview.parse (preview_v);
      assert (reconstructed.title == preview.title);
    }

    static const string ACTIVATE_TEST_URI = "file:///test_uri";
    static const string PREVIEW_TEST_ACTION = "action_id1";

    public void test_scope_preview_action ()
    {
      // we need channel first
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);

      Unity.ScopeResult result = Unity.ScopeResult ();
      result.uri = ACTIVATE_TEST_URI;
      result.icon_hint = "";
      result.category = 0;
      result.result_type = Unity.ResultType.DEFAULT;
      result.mimetype = "inode/folder";
      result.title = "Title";
      result.comment = "";
      result.dnd_uri = "file:///";
      result.metadata = new HashTable<string, Variant> (str_hash, str_equal);
      result.metadata["required_int"] = 5;
      result.metadata["required_string"] = "foo";

      bool got_activate_signal = false;
      scope.activate_action.connect ((result, action_id) =>
      {
        got_activate_signal = true;
        assert (action_id == PREVIEW_TEST_ACTION);
        return new Unity.ActivationResponse (Unity.HandledType.NOT_HANDLED);
      });

      var ml = new MainLoop ();
      Unity.Protocol.ActivationReplyRaw? activation_reply = null;
      var hints = new HashTable<string, Variant> (str_hash, str_equal);
      hints["preview-action-id"] = PREVIEW_TEST_ACTION;
      proxy.activate.begin (channel_id, scope_result_to_variant (result),
                            Unity.Protocol.ActionType.PREVIEW_ACTION,
                            hints, null,
                            (obj, res) =>
      {
        try
        {
          activation_reply = proxy.activate.end (res);
        }
        catch (Error e) {}
        ml.quit ();
      });

      assert (run_with_timeout (ml));
      assert (activation_reply != null);
      assert (activation_reply.handled == Unity.HandledType.NOT_HANDLED);
      assert (got_activate_signal == true);
    }

    public void test_scope_activation ()
    {
      // we need channel first
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);

      Unity.ScopeResult result = Unity.ScopeResult ();
      result.uri = ACTIVATE_TEST_URI;
      result.icon_hint = "";
      result.category = 0;
      result.result_type = Unity.ResultType.DEFAULT;
      result.mimetype = "inode/folder";
      result.title = "Title";
      result.comment = "";
      result.dnd_uri = "file:///";
      result.metadata = new HashTable<string, Variant> (str_hash, str_equal);
      result.metadata["required_int"] = 5;
      result.metadata["required_string"] = "foo";

      perform_search (proxy, channel_id, "", null, null, scope, (search) =>
      {
        search.result_set.add_result (result);
      });

      bool got_activate_signal = false;
      scope.activate_uri.connect ((obj, uri/*, hints*/) =>
      {
        got_activate_signal = true;
        assert (uri == ACTIVATE_TEST_URI);
        return new Unity.ActivationResponse (Unity.HandledType.NOT_HANDLED);
      });

      var ml = new MainLoop ();
      Unity.Protocol.ActivationReplyRaw? activation_reply = null;
      proxy.activate.begin (channel_id, scope_result_to_variant (result),
                            Unity.Protocol.ActionType.ACTIVATE_RESULT,
                            new HashTable<string, Variant> (null, null), null,
                            (obj, res) =>
      {
        try
        {
          activation_reply = proxy.activate.end (res);
        }
        catch (Error e) {}
        ml.quit ();
      });

      assert (run_with_timeout (ml));
      assert (activation_reply != null);
      assert (activation_reply.handled == Unity.HandledType.NOT_HANDLED);
      assert (got_activate_signal == true);
    }

    public void test_scope_filters ()
    {
      SignalWrapper[] signals = {};
      // we need channel first
      var channel_id = open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);

      bool got_search_signal = false;
      signals += new SignalWrapper (scope, scope.search.connect ((search) =>
      {
        got_search_signal = true;

        var filter = search.search_context.filter_state.get_filter_by_id ("check-options") as Unity.OptionsFilter;
        assert (filter.get_option ("option1").active == false);
        assert (filter.get_option ("option2").active == true);
        assert (filter.get_option ("option3").active == false);
      }));

      var state = new HashTable<string, Variant> (str_hash, str_equal);
      var filter = create_filter () as Unity.OptionsFilter;
      filter.get_option ("option2").active = true;
      state["changed-filter-row"] = filter.serialize ();
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", state);

      assert (reply_dict != null);
      assert (got_search_signal);
    }
  }

  class MasterScopeTester: Object, Fixture
  {
    class ChildScope: Unity.DeprecatedScope
    {
      public ChildScope (string dbus_path, string id)
      {
        Object (dbus_path: dbus_path, id: id);
      }

      protected override void constructed ()
      {
        base.constructed ();

        var schema = new Unity.Schema ();
        schema.add_field ("required_string", "s", Unity.Schema.FieldType.REQUIRED);
        schema.add_field ("required_int", "i", Unity.Schema.FieldType.REQUIRED);
        schema.add_field ("optional_string", "s", Unity.Schema.FieldType.OPTIONAL);
        this.schema = schema;
        
        var cats = new Unity.CategorySet ();
        cats.add (new Unity.Category ("1211", "A Category", new GLib.ThemedIcon ("text")));
        cats.add (new Unity.Category ("1991", "Unused category", new GLib.ThemedIcon ("text")));
        categories = cats;
      }
    }

    private uint owning_id;
    private Unity.MasterScope master_scope;
    private ScopeProxy proxy;
    private Unity.DeprecatedScope[] child_scopes = {};
    private Rand random = new Rand ();
    private uint results_per_scope = 1;
    private int random_range_end = 1000000000;
    private bool randomize_categories = false;

    private string[] child_searches = {};
    private int search_handler_invocations
    {
      get
      {
        return child_searches.length;
      }
    }
    private string[] active_filters = {};

    private void child_search_handler (Unity.DeprecatedScope scope,
                                       Unity.DeprecatedScopeSearch search,
                                       Unity.SearchType search_type,
                                       Cancellable canc)
    {
      child_searches += scope.id;

      var filter = search.get_filter ("check-options");
      if (filter != null)
      {
          foreach (var opt in (filter as Unity.CheckOptionFilter).options)
          {
              if (opt.active == true)
                  active_filters += opt.id;
          }
      }

      // simulate async search
      Idle.add (() => { search.finished (); return false; });

      var model = search.results_model;
      var uri = "file:///" + scope.id;
      for (uint i = 0; i < results_per_scope; i++)
      {
        Variant[] row_data;
        uint category_index = 0;
        if (randomize_categories) category_index = random.boolean () ? 1 : 0;
        var rand_int = random.int_range (0, random_range_end);
        if (random.boolean ())
        {
          row_data = model.build_named_row (null,
                                            "uri", uri,
                                            "icon_hint", "",
                                            "category", category_index,
                                            "result_type", 0,
                                            "mimetype", "inode/folder",
                                            "title", "Title",
                                            "comment", scope.id,
                                            "dnd_uri", "file:///",
                                            "required_int", rand_int,
                                            "required_string", "qoo");
        }
        else
        {
          var opt_field = random.next_int ().to_string ();
          row_data = model.build_named_row (null,
                                            "uri", uri,
                                            "icon_hint", "",
                                            "category", category_index,
                                            "result_type", 0,
                                            "mimetype", "inode/folder",
                                            "title", "Title",
                                            "comment", scope.id,
                                            "dnd_uri", "file:///",
                                            "required_int", rand_int,
                                            "required_string", "qoo",
                                            "optional_string", opt_field);
        }
        model.append_row (row_data);
      }
    }

    public static bool own_bus_name (out uint owning_id)
    {
      bool name_owned = false;
      var ml = new MainLoop ();
      // register us a name on the bus
      owning_id = Bus.own_name (BusType.SESSION, DBUS_NAME, 0,
                                () => {},
                                () => { name_owned = true; ml.quit (); },
                                () => { ml.quit (); });
      ml.run ();
      return name_owned;
    }

    private void setup ()
    {
      assert (own_bus_name (out owning_id));
      master_scope = new Unity.MasterScope (MASTER_SCOPE_DBUS_PATH,
                                            "test_masterscope.scope");
      master_scope.search_hint = "Master search hint";
      var schema = new Unity.Schema ();
      schema.add_field ("required_string", "s", Unity.Schema.FieldType.REQUIRED);
      schema.add_field ("required_int", "i", Unity.Schema.FieldType.REQUIRED);
      schema.add_field ("optional_string", "s", Unity.Schema.FieldType.OPTIONAL);
      master_scope.schema = schema;

      var filters = new Unity.FilterSet ();
      filters.add (ScopeTester.create_filter ());
      master_scope.filters = filters;

      var cats = new Unity.CategorySet ();
      cats.add (new Unity.Category ("1991", "Unused category", new GLib.ThemedIcon ("text")));
      cats.add (new Unity.Category ("1211", "A Category", new GLib.ThemedIcon ("text")));
      master_scope.categories = cats;
 
      try
      {
        master_scope.export ();
      }
      catch (Error err) { assert_not_reached (); }

      // init child scopes
      ChildScope child_scope;
      child_scope = new ChildScope ("/com/canonical/unity/scope/childscope_1",
                                    "test_masterscope-childscope_1.scope");
      child_scope.filters = filters; // set filters for one child scope
      child_scope.search_changed.connect (child_search_handler);
      child_scope.export ();
      child_scopes += child_scope;

      child_scope = new ChildScope ("/com/canonical/unity/scope/childscope_2",
                                    "test_masterscope-childscope_2.scope");
      child_scope.search_changed.connect (child_search_handler);
      child_scope.export ();
      child_scopes += child_scope;

      proxy = ScopeInitTester.acquire_test_proxy (MASTER_SCOPE_DBUS_PATH);
      assert (proxy != null);
    }

    private void teardown ()
    {
      ensure_destruction ((owned) proxy);
      master_scope.unexport ();
      ensure_destruction ((owned) master_scope);
      foreach (var scope in child_scopes) scope.unexport ();
      child_scopes = {};
      if (owning_id != 0) Bus.unown_name (owning_id);
    }

    public void test_master_open_channel ()
    {
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);
    }

    public void test_master_open_close_channel ()
    {
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT, null);
      assert (channel_id != null);
      ScopeTester.close_channel (proxy, channel_id);
    }

    public void test_master_search ()
    {
      // we need channel first
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);

      assert (search_handler_invocations == child_scopes.length);
      assert (reply_dict != null);
      assert (model.get_n_rows () > 0);
    }

    public void test_master_activation ()
    {
      // we need channel first
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (model.get_n_rows () >= 2);

      int num_cols = (int) model.get_n_columns ();
      var iter = model.get_first_iter ();
      var result1 = model.get_row (iter, null);
      result1.length = num_cols;
      iter = model.next (iter);
      var result2 = model.get_row (iter, null);
      result2.length = num_cols;

      bool scope1_got_activate = false;
      bool scope2_got_activate = false;
      
      child_scopes[0].activate_uri.connect ((scope, uri) =>
      {
        assert (uri.has_prefix ("file:///") && "childscope_1" in uri);
        scope1_got_activate = true;
        return new Unity.ActivationResponse (Unity.HandledType.NOT_HANDLED);
      });

      child_scopes[1].activate_uri.connect ((scope, uri) =>
      {
        assert (uri.has_prefix ("file:///") && "childscope_2" in uri);
        scope2_got_activate = true;
        return new Unity.ActivationResponse (Unity.HandledType.NOT_HANDLED);
      });

      var ml = new MainLoop ();
      proxy.activate.begin (channel_id, result1,
                            Unity.Protocol.ActionType.ACTIVATE_RESULT,
                            new HashTable<string, Variant> (null, null), null,
                            (obj, res) =>
      {
        ml.quit ();
      });
      assert (run_with_timeout (ml));
      // one of the scope should have received the signal
      assert (scope1_got_activate || scope2_got_activate);
      assert ((scope1_got_activate && scope2_got_activate) == false);

      ml = new MainLoop ();
      proxy.activate.begin (channel_id, result2,
                            Unity.Protocol.ActionType.ACTIVATE_RESULT,
                            new HashTable<string, Variant> (null, null), null,
                            (obj, res) =>
      {
        ml.quit ();
      });
      assert (run_with_timeout (ml));

      assert (scope1_got_activate);
      assert (scope2_got_activate);
    }

    public void test_master_sorting_asc ()
    {
      results_per_scope = 100;
      master_scope.add_sorter (1, "required_int", Unity.AggregatorScope.SortFlags.ASCENDING);
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);
      assert (model.get_n_rows () > 0);

      // only results with one cat were added, check the monotonicity
      bool is_monotonic = true;
      int last_val = int.MIN;
      var iter = model.get_first_iter ();
      var end_iter = model.get_last_iter ();
      while (iter != end_iter)
      {
        var primary_dict = model.get_value (iter, Unity.Internal.ResultColumn.METADATA);
        var content = primary_dict.lookup_value ("content", VariantType.VARDICT);
        int val = content.lookup_value ("required_int", VariantType.INT32).get_int32 ();
        is_monotonic &= val >= last_val;
        last_val = val;
        iter = model.next (iter);
      }

      assert (is_monotonic);
    }

    public void test_master_sorting_desc ()
    {
      results_per_scope = 100;
      master_scope.add_sorter (1, "required_int", Unity.AggregatorScope.SortFlags.DESCENDING);
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);
      assert (model.get_n_rows () > 0);

      // only results with one cat were added, check the monotonicity
      bool is_monotonic = true;
      int last_val = int.MAX;
      var iter = model.get_first_iter ();
      var end_iter = model.get_last_iter ();
      while (iter != end_iter)
      {
        var primary_dict = model.get_value (iter, Unity.Internal.ResultColumn.METADATA);
        var content = primary_dict.lookup_value ("content", VariantType.VARDICT);
        int val = content.lookup_value ("required_int", VariantType.INT32).get_int32 ();
        is_monotonic &= val <= last_val;
        last_val = val;
        iter = model.next (iter);
      }

      assert (is_monotonic);
    }

    public void test_master_sorting_optional_asc ()
    {
      results_per_scope = 100;
      master_scope.add_sorter (1, "optional_string", Unity.AggregatorScope.SortFlags.ASCENDING);
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);

      assert (search_handler_invocations == child_scopes.length);
      assert (model.get_n_rows () > 0);

      // only results with one cat were added, check the monotonicity
      bool is_monotonic = true;
      var iter = model.get_first_iter ();
      var end_iter = model.get_last_iter ();
      // this would spit a critical if first result didn't have the field
      var last_val = model.get_value (iter, Unity.Internal.ResultColumn.METADATA).lookup_value ("content", VariantType.VARDICT).lookup_value ("optional_string", null).get_string ();
      iter = model.next (iter);
      while (iter != end_iter)
      {
        var primary_dict = model.get_value (iter, Unity.Internal.ResultColumn.METADATA);
        var content = primary_dict.lookup_value ("content", VariantType.VARDICT);
        var opt_variant = content.lookup_value ("optional_string", null);
        unowned string val = opt_variant != null ? opt_variant.get_string () : null;
        is_monotonic &= val == null || strcmp (val, last_val) >= 0;
        last_val = val;
        iter = model.next (iter);
      }

      assert (is_monotonic);
    }

    public void test_master_sorting_optional_desc ()
    {
      results_per_scope = 100;
      master_scope.add_sorter (1, "optional_string", Unity.AggregatorScope.SortFlags.DESCENDING);
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);

      assert (search_handler_invocations == child_scopes.length);
      assert (model.get_n_rows () > 0);

      // only results with one cat were added, check the monotonicity
      bool is_monotonic = true;
      var iter = model.get_first_iter ();
      var end_iter = model.get_last_iter ();
      // this would spit a critical if first result didn't have the field
      var last_val = model.get_value (iter, Unity.Internal.ResultColumn.METADATA).lookup_value ("content", VariantType.VARDICT).lookup_value ("optional_string", null).get_string ();
      iter = model.next (iter);
      while (iter != end_iter)
      {
        var primary_dict = model.get_value (iter, Unity.Internal.ResultColumn.METADATA);
        var content = primary_dict.lookup_value ("content", VariantType.VARDICT);
        var opt_variant = content.lookup_value ("optional_string", null);
        unowned string val = opt_variant != null ? opt_variant.get_string () : null;
        is_monotonic &= val == null || strcmp (val, last_val) <= 0;
        last_val = val;
        iter = model.next (iter);
      }

      assert (is_monotonic);
    }

    public void test_master_deduplication ()
    {
      results_per_scope = 100;
      random_range_end = 2;
      master_scope.add_constraint (-1, "required_int");
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);

      assert (search_handler_invocations == child_scopes.length);
      // constraining to only unique values of "required_int" and there are
      // only two of those, even though we tried to add 200 results
      assert (model.get_n_rows () == 2);
    }

    public void test_master_deduplication_per_category ()
    {
      results_per_scope = 100;
      random_range_end = 2;
      randomize_categories = true;

      master_scope.add_constraint (0, "required_int");
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);

      assert (search_handler_invocations == child_scopes.length);
      var counts = new HashTable<uint, uint> (direct_hash, direct_equal);
      var iter = model.get_first_iter ();
      var end_iter = model.get_last_iter ();
      while (iter != end_iter)
      {
        var category = model.get_uint32 (iter, Unity.Internal.ResultColumn.CATEGORY);
        counts[category] = counts[category] + 1;
        iter = model.next (iter);
      }

      // category 0 should have just 2 results (all other are dupes)
      // category 1 should have ~100 (depending on random generator)
      assert (counts[0] == 2);
      assert (counts[1] > 2);
    }

    public void test_master_filters ()
    {
      var expected_error = new ErrorHandler ();
      expected_error.ignore_message ("libunity", LogLevelFlags.LEVEL_WARNING);
      // initialize channel
      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT,
                                                 out model);
      assert (channel_id != null);

      var state = new HashTable<string, Variant> (str_hash, str_equal);
      var filter = ScopeTester.create_filter () as Unity.OptionsFilter;
      filter.get_option ("option2").active = true;
      state["changed-filter-row"] = filter.serialize ();
      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", state, model);

      assert (reply_dict != null);
      // at least one scope should ignore the search
      assert (search_handler_invocations < child_scopes.length);
      assert (model.get_n_rows () > 0);
      assert (active_filters.length == 1);
      assert (active_filters[0] == "option2");
    }

    public void test_subscopes_filter_hint ()
    {
      assert (proxy != null);

      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT, out model);
      assert (channel_id != null);

      var hints = new HashTable<string, Variant> (str_hash, str_equal);
      string[] subscopes = {"test_masterscope-childscope_1.scope"};
      hints ["subscopes_filter"] = new GLib.Variant.strv (subscopes);

      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", hints, model);
      assert (reply_dict != null);

      assert (search_handler_invocations == 1);
      assert (child_searches[0] =="test_masterscope-childscope_1.scope");
    }
  }

  class AggregatorScopeTester: Object, Fixture
  {
    class AggScope: Unity.AggregatorScope
    {
      public AggScope (string dbus_path, string id)
      {
        Object (dbus_path: dbus_path, id: id, is_master: true, merge_mode: Unity.AggregatorScope.MergeMode.OWNER_SCOPE);   
      }

      protected override void constructed ()
      {
        base.constructed ();

        search_hint = "Home search hint";

        var fi = new Unity.FilterSet ();
        fi.add (ScopeTester.create_filter ());
        filters = fi;

        var cats = new Unity.CategorySet ();
        cats.add (new Unity.Category ("test_masterscope.scope", "Masterscope 1", new GLib.ThemedIcon ("text")));
        cats.add (new Unity.Category ("test_fooobar.scope", "Masterscope 2", new GLib.ThemedIcon ("text")));
        categories = cats;
      }

      public override int category_index_for_scope_id (string scope_id)
      {
        return 0;
      }

      public override async void search (Unity.AggregatedScopeSearch scope_search)
      {
        var order = new uint32[] {1,0};
        scope_search.category_order_changed (order);
      }
    }

    private AggScope agg_scope;
    private ScopeProxy proxy;
    private uint owning_id;

    private void setup ()
    {
      assert (MasterScopeTester.own_bus_name (out owning_id));
      // setup Home Scope
      agg_scope = new AggScope (HOME_SCOPE_DBUS_PATH, "test_homescope.scope");
      try
      {
        agg_scope.export ();
      }
      catch (Error err) { assert_not_reached (); }

      proxy = ScopeInitTester.acquire_test_proxy (HOME_SCOPE_DBUS_PATH);
      assert (proxy != null);
    }

    private void teardown ()
    {
      ensure_destruction ((owned) proxy);
      agg_scope.unexport ();
      ensure_destruction ((owned) agg_scope);
      if (owning_id != 0) Bus.unown_name (owning_id);
    }

    public void test_scope_category_order_signal ()
    {
      assert (proxy != null);
      SignalWrapper[] signals = {};
      bool got_order_changed = false;

      signals += new SignalWrapper (proxy, proxy.category_order_changed.connect ((channel, order) =>
      {
        got_order_changed = true;
        assert (order.length == 2);
        assert (order[0] == 1 && order[1] == 0);
      }));

      Dee.SerializableModel model;
      var channel_id = ScopeTester.open_channel (proxy, ChannelType.DEFAULT, out model);
      assert (channel_id != null);
      assert (got_order_changed == false);

      var reply_dict = ScopeTester.perform_search (proxy, channel_id, "", null, model);
      assert (reply_dict != null);
      assert (got_order_changed == true);
    }
  }
}
